import logging

import cfdm

from .constants import (
    formula_terms_computed_standard_names,
    formula_terms_D1,
    formula_terms_max_dimensions,
    formula_terms_standard_names,
    formula_terms_units,
)
from .docstring import _docstring_substitution_definitions
from .functions import bounds_combination_mode
from .units import Units

logger = logging.getLogger(__name__)


class FormulaTerms(metaclass=cfdm.core.DocstringRewriteMeta):
    """Functions for computing non-parametric vertical coordinates from
    the formula defined by a coordinate reference construct.

    {{formula terms links}}

    .. versionaddedd:: 3.8.0

    """

    # The standard names for which there are formulas. There must be a
    # method with the same name.
    standard_names = (
        "atmosphere_ln_pressure_coordinate",
        "atmosphere_sigma_coordinate",
        "atmosphere_hybrid_sigma_pressure_coordinate",
        "atmosphere_hybrid_height_coordinate",
        "atmosphere_sleve_coordinate",
        "ocean_sigma_coordinate",
        "ocean_s_coordinate",
        "ocean_s_coordinate_g1",
        "ocean_s_coordinate_g2",
        "ocean_sigma_z_coordinate",
        "ocean_double_sigma_coordinate",
    )

    def __docstring_substitutions__(self):
        """Define docstring substitutions that apply to this class and
        all of its subclasses.

        These are in addtion to, and take precendence over, docstring
        substitutions defined by the base classes of this class.

        See `_docstring_substitutions` for details.

        .. versionadded:: 3.8.0

        .. seealso:: `_docstring_substitutions`

        :Returns:

            `dict`
                The docstring substitutions that have been applied.

        """
        return _docstring_substitution_definitions

    def __docstring_package_depth__(self):
        """Return the package depth for {{package}} docstring
        substitutions.

        See `_docstring_package_depth` for details.

        """
        return 1

    # ----------------------------------------------------------------
    # Private methods
    # ----------------------------------------------------------------
    @staticmethod
    def _domain_ancillary_term(
        f,
        standard_name,
        coordinate_conversion,
        term,
        default_to_zero,
        strict,
        bounds,
    ):
        """Find a domain ancillary construct corresponding to a formula
        term.

        .. versionadded:: 3.8.0

        :Parameters:

            f: `Field`
                The parent field construct.

            standard_name: `str`
                The standard name of the parametric vertical coordinate.

            coordinate_conversion: `CoordinateConversion`
                The definition of the formula

            term: `str`
                A term of the formula.

                *Parameter example:*
                  ``term='orog'``

            {{default_to_zero: `bool`, optional}}

            strict: `bool`
                If False then allow the computation to occur when

                * A term with a defined standard name (or names) has no
                  standard name.

                * When the computed standard name can not be found by
                  inference from the standard names of the domain
                  ancillary constructs; or from the
                  ``computed_standard_name`` property.

            bounds: `bool`, optional
                If False then do not create bounds of zero for a default
                domain ancillary construct.

        :Returns:

            `DomainAncillary`, `str`
                The domain ancillary construct for the formula term and
                its construct key.

        """
        units = formula_terms_units[standard_name].get(term, None)
        if units is None:
            raise ValueError(
                "Can't calculate non-parametric vertical coordinates: "
                f"{term!r} is not a valid term"
            )

        units = Units(units)

        key = coordinate_conversion.get_domain_ancillary(term, None)
        if key is not None:
            var = f.domain_ancillary(key, default=None)
        else:
            var = None

        if var is not None:
            logger.detail(
                f"Formula term {term!r}:\n{var.dump(display=False, _level=1)}"
            )  # pragma: no cover

            valid_standard_names = formula_terms_standard_names[standard_name][
                term
            ]

            vsn = var.get_property("standard_name", None)

            if vsn is None:
                if (
                    strict
                    and vsn not in valid_standard_names
                    and len(valid_standard_names) > 1
                ):
                    expected_names = ", ".join(
                        repr(x) for x in valid_standard_names
                    )
                    raise ValueError(
                        "Can't calculate non-parametric vertical coordinates: "
                        f"{term!r} term {var!r} has no standard name. "
                        f"Expected one of {expected_names}"
                    )
            elif vsn not in valid_standard_names:
                expected_names = ", ".join(
                    repr(x) for x in valid_standard_names
                )
                raise ValueError(
                    "Can't calculate non-parametric vertical coordinates: "
                    f"{term!r} term {var!r} has invalid standard name: "
                    f"{vsn!r}. Expected one of {expected_names}"
                )

            if var.ndim > formula_terms_max_dimensions[standard_name][term]:
                raise ValueError(
                    "Can't calculate non-parametric vertical coordinates: "
                    f"{term!r} term {var!r} has incorrect "
                    f"number of dimensions. Expected at most {var.ndim}"
                )

            if not var.Units.equivalent(units):
                raise ValueError(
                    "Can't calculate non-parametric vertical coordinates: "
                    f"{term!r} term {var!r} has incorrect units: "
                    f"{var.Units!r}. Expected units equivalent to {units!r}"
                )
        else:
            if not default_to_zero:
                raise ValueError(
                    "Can't calculate non-parametric vertical coordinates: "
                    f"No {term!r} term domain ancillary construct and "
                    "default_to_zero=False"
                )
            # ----------------------------------------------------
            # Create a default zero-valued domain ancillary
            # ----------------------------------------------------
            var = f._DomainAncillary()

            data = f._Data(0.0, units)
            var.set_data(data)

            if bounds:
                bounds = f._Bounds(data=f._Data((0.0, 0.0), units))
                var.set_bounds(bounds)

            logger.detail(
                f"Formula term {term!r} (default):\n"
                f"{var.dump(display=False, _level=1)}"
            )  # pragma: no cover

        return var, key

    @staticmethod
    def _computed_standard_name(f, standard_name, coordinate_reference):
        """Find the standard name of the computed non-parametric
        vertical coordinates.

        {{formula terms links}}

        .. versionadded:: 3.8.0

        :Parameters:

            f: `Field`
                The parent field construct.

            standard_name: `str`
                The standard name of the parametric vertical coordinate.

            coordinate_reference: `CoordinateReference`
                A coordinate reference construct of the parent field
                construct.

        :Returns:

            `str` or `None`
                The standard name of the computed vertical coordinates, or
                `None` if one could not be found.

        """
        computed_standard_name = formula_terms_computed_standard_names[
            standard_name
        ]

        if isinstance(computed_standard_name, str):
            # ------------------------------------------------------------
            # There is a unique computed standard name for this formula
            # ------------------------------------------------------------
            logger.detail(
                f"computed_standard_name: {computed_standard_name!r}"
            )  # pragma: no cover

            return computed_standard_name

        # ----------------------------------------------------------------
        # Still here? Then see if the computed standard name depends on
        # the standard name of one of the formula terms.
        # ----------------------------------------------------------------
        term, mapping = tuple(computed_standard_name.items())[0]

        computed_standard_name = None

        coordinate_conversion = coordinate_reference.coordinate_conversion
        key = coordinate_conversion.get_domain_ancillary(term, None)

        if key is not None:
            var = f.domain_ancillary(key, default=None)
            if var is not None:
                term_standard_name = var.get_property("standard_name", None)
                if term_standard_name is not None:
                    for x, y in mapping.items():
                        if term_standard_name == x:
                            computed_standard_name = y
                            break

        if computed_standard_name is None:
            # ------------------------------------------------------------
            # See if the computed standard name is a set as a coordinate
            # conversion parameter
            # ------------------------------------------------------------
            computed_standard_name = coordinate_conversion.get_parameter(
                "computed_standard_name", None
            )

        if computed_standard_name is None:
            # ------------------------------------------------------------
            # As a last resort use the computed_standard_name property of
            # the parametric vertical coordinate construct
            # ------------------------------------------------------------
            var = None
            key = None

            for key in coordinate_reference.coordinates():
                var = f.coordinate(key, default=None)
                if var is None:
                    continue

                if var.get_property("standard_name", None) != standard_name:
                    continue

                computed_standard_name = var.get_property(
                    "computed_standard_name", None
                )
                break

        logger.detail(
            "computed_standard_name: {}".format(
                repr(computed_standard_name) if computed_standard_name else ""
            )
        )  # pragma: no cover

        return computed_standard_name

    @staticmethod
    def _vertical_axis(f, keys):
        """Find the vertical axis corresponding to the parametric
        vertical coordinates.

        .. versionadded:: 3.8.0

        :Parameters:

            f: `Field`
                The parent field construct.

            keys: sequence of `str` or `None`
                The construct keys of 1-d domain ancillary or coordinate
                constructs that span the vertical axis. If a key is `None`
                then that key is ignored.

        :Returns:

            `tuple`
                Either a 1-tuple containing the domain axis construct key
                of the vertical axis, or an empty tuple if no such axis
                could be found.

        """
        axis = ()
        for key in keys:
            if key is None:
                continue

            axis = f.get_data_axes(key)
            break

        logger.detail(f"Vertical axis: {axis!r}")  # pragma: no cover

        return axis

    @staticmethod
    def _conform_eta(f, eta, eta_key, depth, depth_key):
        """Transform the 'eta' term so that broadcasting will work with
        the 'depth' term.

        This entails making sure that the trailing dimensions of 'eta' are
        the same as all of the dimensions of 'depth'.

        For example, if we have ``eta(i,n,j)`` and ``depth(j,i)`` then eta
        will be transformed to ``eta(n,j,i)``.

        .. versionadded:: 3.8.0

        :Parameters:

            f: `Field`
                The parent field construct.

            eta: `DomainAncillary`
                The 'eta' domain ancillary construct.

            eta_key: `str`
                The construct key of the 'eta' domain ancillary construct.

            depth: `DomainAncillary`
                The 'depth' domain ancillary construct.

            depth_key: `str`
                The construct key of the 'depth' domain ancillary
                construct.

        :Returns:

            `DomainAncillary`, `tuple`
                The conformed 'eta' domain ancillary construct for the
                formula term, and the domain axis construct keys of its
                dimensions.

        """
        eta_axes = f.get_data_axes(eta_key, default=None)
        depth_axes = f.get_data_axes(depth_key, default=None)

        if eta_axes is not None and depth_axes is not None:
            if not set(eta_axes).issuperset(depth_axes):
                raise ValueError(
                    "Can't calculate non-parametric coordinates: "
                    f"'depth' term {depth!r} axes must be a subset of "
                    f"'eta' term {eta!r} axes."
                )

            eta_axes2 = depth_axes
            if len(eta_axes) > len(depth_axes):
                diff = [axis for axis in eta_axes if axis not in depth_axes]
                eta_axes2 = tuple(diff) + depth_axes

            iaxes = [eta_axes.index(axis) for axis in eta_axes2]
            eta = eta.transpose(iaxes)

            eta_axes = eta_axes2

        logger.debug(
            f"Transposed domain ancillary 'eta': {eta!r}\n"
            f"Transposed domain ancillary 'eta' axes: {eta_axes!r}"
        )  # pragma: no cover

        return eta, eta_axes

    @staticmethod
    def _conform_computed(f, computed, computed_axes, k_axis):
        """Move the vertical axis of the computed non-parametric
        vertical coordinates from its current position as the last
        (rightmost) dimension, if applicable.

        **Note:**

        * If the input computed coordinates do not span the vertical axis
          then they are returned unchanged.

        * If the input computed coordinates do span the vertical axis then
          it assumed to be the last (rightmost) dimension.

        * If the input computed coordinates span a unique time axis then
          the vertical axis is moved to position immediately to the right
          of it. Otherwise the vertical axis is moved to the first
          (leftmost) position.

        .. versionadded:: 3.8.0

        :Parameters:

            f: `Field`
                The parent field construct.

            computed: `DomainAncillary`
                The computed non-parametric vertical coordinates.

            computed_axes: `tuple`
                The construct keys of the domain axes spanned by the
                computed coordinates.

            k_axis: `tuple`
                The construct key of the vertical domain axis. If the
                vertical axis does not appear in the computed
                non-parametric coodinates then this must be an empty
                tuple.

        :Returns:

            `DomainAncillary`, `tuple`
                The conformed computed non-parametric vertical coordinates
                construct and the domain axis construct keys of its
                dimensions.

        """
        ndim = computed.ndim
        if k_axis and ndim >= 2:
            iaxes = list(range(ndim - 1))

            time = f.domain_axis("T", key=True, default=None)
            if time in computed_axes:
                # Move it to the immediate right of the time
                # axis
                iaxes.insert(computed_axes.index(time) + 1, -1)
            else:
                # Move to to position 0, as there is no time
                # axis.
                iaxes.insert(0, -1)

            computed.transpose(iaxes, inplace=True)

            computed_axes = tuple([computed_axes[i] for i in iaxes])

        logger.detail(
            f"Non-parametric coordinate axes: {computed_axes!r}"
        )  # pragma: no cover

        return computed, computed_axes

    @staticmethod
    def _conform_units(term, var, ref_term2, ref_units):
        """Make sure that the units of a variable of the formula are the
        same as the units of another formula term.

        .. versionadded:: 3.8.0

        :Parameters:

            term: `str`
                A term of the formula for which the units are to be
                conformed.

                *Parameter example:*
                  ``term='z2'``

            var: `DomainAncillary` or `Coordinate`
                The variable for the *term* term whose units are to
                conformed to *ref_units*.

            ref_term: `str`
                A term of the formula which defines the reference units.

                *Parameter example:*
                  ``ref_term='href'``

            ref_units: `Units`
                The units of the *ref_term* term.

                *Parameter example:*
                  ``ref_units=cf.Units('1')`

        :Returns:

            `DomainAncillary` or `Coordinate`
                The input *var* construct with conformed units.

        """
        units = var.Units

        if units != ref_units:
            if not units.equivalent(ref_units):
                raise ValueError(
                    f"Terms {ref_term2!r} and {term!r} have incompatible "
                    f"units: {ref_units!r}, {units!r}"
                )

            var = var.copy()
            var.Units = ref_units

        return var

    @staticmethod
    def _check_standard_name_consistency(
        strict, zlev=(None, None), eta=(None, None), depth=(None, None)
    ):
        """Check that there are consistent sets of values for the
        standard_names of formula terms and the computed_standard_name
        needed in defining the ocean sigma coordinate, the ocean
        s-coordinate, the ocean_sigma over z coordinate, and the ocean
        double sigma coordinate.

        {{formula terms links}}

        .. versionadded:: 3.8.0

        :Parameters:

            zlev: `tuple`, optional
                The 'zlev' term domain ancillary construct and its
                construct key. If the domain ancillary construct was
                created from default values then the construct key will be
                `None`.

            eta: `tuple`, optional
                The 'eta' term domain ancillary construct and its
                construct key. If the domain ancillary construct was
                created from default values then the construct key will be
                `None`.

            depth: `tuple`, optional
                The 'depth' term domain ancillary construct and its
                construct key. If the domain ancillary construct was
                created from default values then the construct key will be
                `None`.

            strict: `bool`
                If False then allow the computation to occur even if none
                of the given the 'zlev', 'eta', and 'depth' domain
                ancillary constructs have a standard name. By default if
                any of these constructs is missing a standard name then an
                exception will be raised.

        :Returns:

            `None`
                A `ValueError` is raised if the standard names are
                inconsistent.

        """
        kwargs = locals()
        kwargs.pop("strict")

        indices = set()

        for term, (var, key) in kwargs.items():
            if var is None:
                continue

            standard_name = var.get_property("standard_name", None)

            if standard_name is None and key is None:
                continue

            try:
                indices.add(formula_terms_D1[term].index(standard_name))
            except ValueError:
                # Note that the existence of a standard name has already
                # been ensured by `_domain_ancillary_term`
                if strict:
                    raise ValueError(
                        f"{term!r} term {var!r} has invalid "
                        f"standard name: {standard_name!r}"
                    )

        if strict and not indices:
            raise ValueError(
                "Terms {} have no standard names. "
                "See Appendix D: Parametric Vertical Coordinates "
                "of the CF conventions.".format(
                    ", ".join(repr(term) for term in kwargs),
                )
            )

        if len(indices) > 1:
            raise ValueError(
                "Terms {} have incompatible standard names. "
                "See Appendix D: Parametric Vertical Coordinates "
                "of the CF conventions.".format(
                    ", ".join(repr(term) for term in kwargs),
                )
            )

    @staticmethod
    def _check_index_term(term, var):
        """Check that an index term contains a single integer.

        .. versionadded:: 3.8.0

        :Parameters:

            term: `str`
                A term of the formula.

                *Parameter example:*
                  ``term='nsigma'``

            eta: `DomainAncillary`
                The corresponding domain ancillary construct.

        :Returns:

            `None`
                A `ValueEerror` is raised if the index term does not
                contain a single integer.

        """
        if var.size != 1 or var.dtype.kind != "i":
            raise ValueError(
                "Can't calculate non-parametric vertical coordinates: "
                f"{term!r} term {var!r} doen't contain exactly one "
                "integer value"
            )

    # ----------------------------------------------------------------
    # Methods
    # ----------------------------------------------------------------
    @classmethod
    def atmosphere_ln_pressure_coordinate(
        cls, g, coordinate_reference, default_to_zero, strict
    ):
        """Compute non-parametric vertical coordinates from
        atmosphere_ln_pressure_coordinate parametric coordinates.

        {{formula terms links}}

        .. note:: The vertical axis is the last (rightmost) dimension of
                  the returned computed non-parametric vertical
                  coordinates, if applicable.

        .. versionadded:: 3.8.0

        :Parameters:

            f: `Field`
                The parent field construct.

            coordinate_reference: `CoordinateReference`
                The coordinate reference construct of the parent field
                construct that defines the conversion formula.

            {{default_to_zero: `bool`, optional}}

        :Returns:

            {{Returns formula}}

        """
        standard_name = "atmosphere_ln_pressure_coordinate"

        computed_standard_name = cls._computed_standard_name(
            g, standard_name, coordinate_reference
        )

        # ----------------------------------------------------------------
        # Get the formula terms and their contruct keys
        # ----------------------------------------------------------------
        coordinate_conversion = coordinate_reference.coordinate_conversion

        lev, lev_key = cls._domain_ancillary_term(
            g,
            standard_name,
            coordinate_conversion,
            "lev",
            default_to_zero,
            strict,
            True,
        )

        p0, _ = cls._domain_ancillary_term(
            g,
            standard_name,
            coordinate_conversion,
            "p0",
            default_to_zero,
            strict,
            True,
        )

        # Get the axes of the non-parametric coordinates, putting the
        # vertical axis in postition -1 (the rightmost position).
        k_axis = cls._vertical_axis(g, [lev_key])
        computed_axes = k_axis

        # ----------------------------------------------------------------
        # Compute the non-parametric coordinates
        # ----------------------------------------------------------------
        with bounds_combination_mode("OR" if lev.has_bounds() else "AND"):
            computed = p0 * (-lev).exp()

        return (
            standard_name,
            computed_standard_name,
            computed,
            computed_axes,
            k_axis,
        )

    @classmethod
    def atmosphere_sigma_coordinate(
        cls, g, coordinate_reference, default_to_zero, strict
    ):
        """Compute non-parametric vertical coordinates from
        atmosphere_sigma_coordinate parametric coordinates.

        {{formula terms links}}

        .. note:: The vertical axis is the last (rightmost) dimension of
                  the returned computed non-parametric vertical
                  coordinates, if applicable.

        .. versionadded:: 3.8.0

        :Parameters:

            f: `Field`
                The parent field construct.

            coordinate_reference: `CoordinateReference`
                The coordinate reference construct of the parent field
                construct that defines the conversion formula.

            {{default_to_zero: `bool`, optional}}

        :Returns:

            {{Returns formula}}

        """
        standard_name = "atmosphere_sigma_coordinate"

        computed_standard_name = cls._computed_standard_name(
            g, standard_name, coordinate_reference
        )

        # ----------------------------------------------------------------
        # Get the formula terms and their contruct keys
        # ----------------------------------------------------------------
        coordinate_conversion = coordinate_reference.coordinate_conversion

        sigma, sigma_key = cls._domain_ancillary_term(
            g,
            standard_name,
            coordinate_conversion,
            "sigma",
            default_to_zero,
            strict,
            True,
        )

        ptop, _ = cls._domain_ancillary_term(
            g,
            standard_name,
            coordinate_conversion,
            "ptop",
            default_to_zero,
            strict,
            True,
        )

        ps, ps_key = cls._domain_ancillary_term(
            g,
            standard_name,
            coordinate_conversion,
            "ps",
            default_to_zero,
            strict,
            True,
        )

        # Get the axes of the non-parametric coordinates, putting the
        # vertical axis in postition -1 (the rightmost position).
        k_axis = cls._vertical_axis(g, [sigma_key])
        computed_axes = g.get_data_axes(ps_key) + k_axis

        # Insert a size one dimension to allow broadcasting
        # over the vertical axis
        if k_axis:
            ps = ps.insert_dimension(-1)

        # ----------------------------------------------------------------
        # Compute the non-parametric coordinates
        # ----------------------------------------------------------------
        with bounds_combination_mode("OR" if sigma.has_bounds() else "AND"):
            computed = ptop + (ps - ptop) * sigma

        return (
            standard_name,
            computed_standard_name,
            computed,
            computed_axes,
            k_axis,
        )

    @classmethod
    def atmosphere_hybrid_sigma_pressure_coordinate(
        cls, g, coordinate_reference, default_to_zero, strict
    ):
        """Compute non-parametric vertical coordinates from
        atmosphere_hybrid_sigma_pressure_coordinate parametric
        coordinates.

        {{formula terms links}}

        .. note:: The vertical axis is the last (rightmost) dimension of
                  the returned computed non-parametric vertical
                  coordinates, if applicable.

        .. versionadded:: 3.8.0

        :Parameters:

            f: `Field`
                The parent field construct.

            coordinate_reference: `CoordinateReference`
                The coordinate reference construct of the parent field
                construct that defines the conversion formula.

            {{default_to_zero: `bool`, optional}}

        :Returns:

            {{Returns formula}}

        """
        standard_name = "atmosphere_hybrid_sigma_pressure_coordinate"

        computed_standard_name = cls._computed_standard_name(
            g, standard_name, coordinate_reference
        )

        # ----------------------------------------------------------------
        # Get the formula terms and their contruct keys
        # ----------------------------------------------------------------
        coordinate_conversion = coordinate_reference.coordinate_conversion

        ap_term = "ap" in coordinate_conversion.domain_ancillaries()

        if ap_term:
            ap, ap_key = cls._domain_ancillary_term(
                g,
                standard_name,
                coordinate_conversion,
                "ap",
                default_to_zero,
                strict,
                True,
            )

            a_key = None
        else:
            a, a_key = cls._domain_ancillary_term(
                g,
                standard_name,
                coordinate_conversion,
                "a",
                default_to_zero,
                strict,
                True,
            )

            p0, _ = cls._domain_ancillary_term(
                g,
                standard_name,
                coordinate_conversion,
                "p0",
                default_to_zero,
                strict,
                False,
            )

            ap_key = None

        b, b_key = cls._domain_ancillary_term(
            g,
            standard_name,
            coordinate_conversion,
            "b",
            default_to_zero,
            strict,
            True,
        )

        ps, ps_key = cls._domain_ancillary_term(
            g,
            standard_name,
            coordinate_conversion,
            "ps",
            default_to_zero,
            strict,
            False,
        )

        # Get the axes of the non-parametric coordinates, putting the
        # vertical axis in postition -1 (the rightmost position).
        k_axis = cls._vertical_axis(g, [ap_key, a_key, b_key])
        computed_axes = g.get_data_axes(ps_key) + k_axis

        # Insert a size one dimension to allow broadcasting over the
        # vertical axis
        if k_axis:
            ps = ps.insert_dimension(-1)

        # ----------------------------------------------------------------
        # Compute the non-parametric coordinates
        # ----------------------------------------------------------------
        if ap_term:
            with bounds_combination_mode(
                "OR" if ap.has_bounds() and b.has_bounds() else "AND"
            ):
                computed = ap + b * ps
        else:
            with bounds_combination_mode(
                "OR" if a.has_bounds() and b.has_bounds() else "AND"
            ):
                computed = a * p0 + b * ps

        return (
            standard_name,
            computed_standard_name,
            computed,
            computed_axes,
            k_axis,
        )

    @classmethod
    def atmosphere_hybrid_height_coordinate(
        cls, g, coordinate_reference, default_to_zero, strict
    ):
        """Compute non-parametric vertical coordinates from
        atmosphere_hybrid_height_coordinate parametric coordinates.

        {{formula terms links}}

        .. note:: The vertical axis is the last (rightmost) dimension of
                  the returned computed non-parametric vertical
                  coordinates, if applicable.

        .. versionadded:: 3.8.0

        :Parameters:

            f: `Field`
                The parent field construct.

            coordinate_reference: `CoordinateReference`
                The coordinate reference construct of the parent field
                construct that defines the conversion formula.

            {{default_to_zero: `bool`, optional}}

        :Returns:

            {{Returns formula}}

        """
        standard_name = "atmosphere_hybrid_height_coordinate"

        computed_standard_name = cls._computed_standard_name(
            g, standard_name, coordinate_reference
        )

        # ----------------------------------------------------------------
        # Get the formula terms and their contruct keys
        # ----------------------------------------------------------------
        coordinate_conversion = coordinate_reference.coordinate_conversion

        a, a_key = cls._domain_ancillary_term(
            g,
            standard_name,
            coordinate_conversion,
            "a",
            default_to_zero,
            strict,
            True,
        )

        b, b_key = cls._domain_ancillary_term(
            g,
            standard_name,
            coordinate_conversion,
            "b",
            default_to_zero,
            strict,
            True,
        )

        orog, orog_key = cls._domain_ancillary_term(
            g,
            standard_name,
            coordinate_conversion,
            "orog",
            default_to_zero,
            strict,
            False,
        )

        # Get the axes of the non-parametric coordinates, putting the
        # vertical axis in postition -1 (the rightmost position).
        k_axis = cls._vertical_axis(g, [a_key, b_key])
        computed_axes = g.get_data_axes(orog_key) + k_axis

        # Insert a size one dimension to allow broadcasting over the
        # vertical axis
        if k_axis:
            orog = orog.insert_dimension(-1)

        # ----------------------------------------------------------------
        # Compute the non-parametric coordinates
        # ----------------------------------------------------------------
        with bounds_combination_mode(
            "OR" if a.has_bounds() and b.has_bounds() else "AND"
        ):
            computed = a + b * orog

        return (
            standard_name,
            computed_standard_name,
            computed,
            computed_axes,
            k_axis,
        )

    @classmethod
    def atmosphere_sleve_coordinate(
        cls, g, coordinate_reference, default_to_zero, strict
    ):
        """Compute non-parametric vertical coordinates from
        atmosphere_sleve_coordinate parametric coordinates.

        {{formula terms links}}

        .. note:: The vertical axis is the last (rightmost) dimension of
                  the returned computed non-parametric vertical
                  coordinates, if applicable.

        .. versionadded:: 3.8.0

        :Parameters:

            f: `Field`
                The parent field construct.

            coordinate_reference: `CoordinateReference`
                The coordinate reference construct of the parent field
                construct that defines the conversion formula.

            {{default_to_zero: `bool`, optional}}

        :Returns:

            {{Returns formula}}

        """
        standard_name = "atmosphere_sleve_coordinate"

        computed_standard_name = cls._computed_standard_name(
            g, standard_name, coordinate_reference
        )

        # ----------------------------------------------------------------
        # Get the formula terms and their contruct keys
        # ----------------------------------------------------------------
        coordinate_conversion = coordinate_reference.coordinate_conversion

        a, a_key = cls._domain_ancillary_term(
            g,
            standard_name,
            coordinate_conversion,
            "a",
            default_to_zero,
            strict,
            True,
        )

        b1, b1_key = cls._domain_ancillary_term(
            g,
            standard_name,
            coordinate_conversion,
            "b1",
            default_to_zero,
            strict,
            True,
        )

        b2, b2_key = cls._domain_ancillary_term(
            g,
            standard_name,
            coordinate_conversion,
            "b2",
            default_to_zero,
            strict,
            True,
        )

        ztop, _ = cls._domain_ancillary_term(
            g,
            standard_name,
            coordinate_conversion,
            "ztop",
            default_to_zero,
            strict,
            False,
        )

        zsurf1, zsurf1_key = cls._domain_ancillary_term(
            g,
            standard_name,
            coordinate_conversion,
            "zsurf1",
            default_to_zero,
            strict,
            False,
        )

        zsurf2, zsurf2_key = cls._domain_ancillary_term(
            g,
            standard_name,
            coordinate_conversion,
            "zsurf2",
            default_to_zero,
            strict,
            False,
        )

        # Make sure that zsurf1 and zsurf2 have the same number of axes
        # and the same axis order
        zsurf1_axes = g.get_data_axes(zsurf1_key, default=None)
        zsurf2_axes = g.get_data_axes(zsurf2_key, default=None)

        if zsurf1_axes is not None and zsurf2_axes is not None:
            if set(zsurf1_axes) != set(zsurf2_axes):
                raise ValueError(
                    "Can't calculate non-parametric coordinates: "
                    "'zsurf1' and 'zsurf2' terms "
                    "domain ancillaries span different domain axes"
                )

            iaxes = [zsurf2_axes.index(axis) for axis in zsurf1_axes]
            zsurf2 = zsurf2.transpose(iaxes)

            zsurf_axes = zsurf1_axes
        elif zsurf1_axes is not None:
            zsurf_axes = zsurf1_axes
        elif zsurf1_axes is not None:
            zsurf_axes = zsurf2_axes
        else:
            zsurf_axes = ()

        # Get the axes of the non-parametric coordinates, putting the
        # vertical axis in postition -1 (the rightmost position).
        k_axis = cls._vertical_axis(g, [a_key, b1_key, b2_key])
        computed_axes = zsurf_axes + k_axis

        # Insert a size one dimension to allow broadcasting over the
        # vertical axis
        if k_axis:
            zsurf1 = zsurf1.insert_dimension(-1)
            zsurf2 = zsurf2.insert_dimension(-1)

        # ----------------------------------------------------------------
        # Compute the non-parametric coordinates
        # ----------------------------------------------------------------
        with bounds_combination_mode(
            "OR"
            if a.has_bounds() and b1.has_bounds() and b2.has_bounds()
            else "AND"
        ):
            computed = a * ztop + b1 * zsurf1 + b2 * zsurf2

        return (
            standard_name,
            computed_standard_name,
            computed,
            computed_axes,
            k_axis,
        )

    @classmethod
    def ocean_sigma_coordinate(
        cls, g, coordinate_reference, default_to_zero, strict
    ):
        """Compute non-parametric vertical coordinates from
        ocean_sigma_coordinate parametric coordinates.

        {{formula terms links}}

        .. note:: The vertical axis is the last (rightmost) dimension of
                  the returned computed non-parametric vertical
                  coordinates, if applicable.

        .. versionadded:: 3.8.0

        :Parameters:

            f: `Field`
                The parent field construct.

            coordinate_reference: `CoordinateReference`
                The coordinate reference construct of the parent field
                construct that defines the conversion formula.

            {{default_to_zero: `bool`, optional}}

        :Returns:

            {{Returns formula}}

        """
        standard_name = "ocean_sigma_coordinate"

        computed_standard_name = cls._computed_standard_name(
            g, standard_name, coordinate_reference
        )

        # ----------------------------------------------------------------
        # Get the formula terms and their contruct keys
        # ----------------------------------------------------------------
        coordinate_conversion = coordinate_reference.coordinate_conversion

        sigma, sigma_key = cls._domain_ancillary_term(
            g,
            standard_name,
            coordinate_conversion,
            "sigma",
            default_to_zero,
            strict,
            True,
        )

        eta, eta_key = cls._domain_ancillary_term(
            g,
            standard_name,
            coordinate_conversion,
            "eta",
            default_to_zero,
            strict,
            False,
        )

        depth, depth_key = cls._domain_ancillary_term(
            g,
            standard_name,
            coordinate_conversion,
            "depth",
            default_to_zero,
            strict,
            False,
        )

        cls._check_standard_name_consistency(
            strict, eta=(eta, eta_key), depth=(depth, depth_key)
        )

        # Make sure that eta and depth are consistent, which may require
        # eta to be transposed.
        eta, eta_axes = cls._conform_eta(g, eta, eta_key, depth, depth_key)

        # Get the axes of the non-parametric coordinates, putting the
        # vertical axis in postition -1 (the rightmost position).
        k_axis = cls._vertical_axis(g, [sigma_key])
        computed_axes = eta_axes + k_axis

        # Insert a size one dimension to allow broadcasting over the
        # vertical axis
        if k_axis:
            eta = eta.insert_dimension(-1)
            depth = depth.insert_dimension(-1)

        # ----------------------------------------------------------------
        # Compute the non-parametric coordinates
        # ----------------------------------------------------------------
        with bounds_combination_mode("OR" if sigma.has_bounds() else "AND"):
            computed = eta + (eta + depth) * sigma

        return (
            standard_name,
            computed_standard_name,
            computed,
            computed_axes,
            k_axis,
        )

    @classmethod
    def ocean_s_coordinate(
        cls, g, coordinate_reference, default_to_zero, strict
    ):
        """Compute non-parametric vertical coordinates from
        ocean_s_coordinate parametric coordinates.

        {{formula terms links}}

        .. note:: The vertical axis is the last (rightmost) dimension of
                  the returned computed non-parametric vertical
                  coordinates, if applicable.

        .. versionadded:: 3.8.0

        :Parameters:

            f: `Field`
                The parent field construct.

            coordinate_reference: `CoordinateReference`
                The coordinate reference construct of the parent field
                construct that defines the conversion formula.

            {{default_to_zero: `bool`, optional}}

        :Returns:

            {{Returns formula}}

        """
        standard_name = "ocean_s_coordinate"

        computed_standard_name = cls._computed_standard_name(
            g, standard_name, coordinate_reference
        )

        # ----------------------------------------------------------------
        # Get the formula terms and their construct keys
        # ----------------------------------------------------------------
        coordinate_conversion = coordinate_reference.coordinate_conversion

        s, s_key = cls._domain_ancillary_term(
            g,
            standard_name,
            coordinate_conversion,
            "s",
            default_to_zero,
            strict,
            True,
        )

        eta, eta_key = cls._domain_ancillary_term(
            g,
            standard_name,
            coordinate_conversion,
            "eta",
            default_to_zero,
            strict,
            False,
        )

        depth, depth_key = cls._domain_ancillary_term(
            g,
            standard_name,
            coordinate_conversion,
            "depth",
            default_to_zero,
            strict,
            False,
        )

        a, _ = cls._domain_ancillary_term(
            g,
            standard_name,
            coordinate_conversion,
            "a",
            default_to_zero,
            strict,
            True,
        )

        b, _ = cls._domain_ancillary_term(
            g,
            standard_name,
            coordinate_conversion,
            "b",
            default_to_zero,
            strict,
            True,
        )

        depth_c, _ = cls._domain_ancillary_term(
            g,
            standard_name,
            coordinate_conversion,
            "depth_c",
            default_to_zero,
            strict,
            False,
        )

        cls._check_standard_name_consistency(
            strict, eta=(eta, eta_key), depth=(depth, depth_key)
        )

        # Make sure that eta and depth are consistent, which may require
        # eta to be transposed.
        eta, eta_axes = cls._conform_eta(g, eta, eta_key, depth, depth_key)

        # Get the axes of the non-parametric coordinates, putting the
        # vertical axis in postition -1 (the rightmost position).
        k_axis = cls._vertical_axis(g, [s_key])
        computed_axes = eta_axes + k_axis

        # Insert a size one dimension to allow broadcasting over the
        # vertical axis
        if k_axis:
            eta = eta.insert_dimension(-1)
            depth = depth.insert_dimension(-1)

        # ----------------------------------------------------------------
        # Compute the non-parametric coordinates as
        # ----------------------------------------------------------------
        # Ensure that a has the same units as s
        a = cls._conform_units("a", a, "s", s.Units)

        with bounds_combination_mode("OR" if s.has_bounds() else "AND"):
            C = (1 - b) * (a * s).sinh() / a.sinh() + b * (
                (a * (s + 0.5)).tanh() / (2 * (a * 0.5).tanh()) - 0.5
            )

            computed = eta * (s + 1) + depth_c * s + (depth - depth_c) * C

        return (
            standard_name,
            computed_standard_name,
            computed,
            computed_axes,
            k_axis,
        )

    @classmethod
    def ocean_s_coordinate_g1(
        cls, g, coordinate_reference, default_to_zero, strict
    ):
        """Compute non-parametric vertical coordinates from
        ocean_s_coordinate_g1 parametric coordinates.

        {{formula terms links}}

        .. note:: The vertical axis is the last (rightmost) dimension of
                  the returned computed non-parametric vertical
                  coordinates, if applicable.

        .. versionadded:: 3.8.0

        :Parameters:

            f: `Field`
                The parent field construct.

            coordinate_reference: `CoordinateReference`
                The coordinate reference construct of the parent field
                construct that defines the conversion formula.

            {{default_to_zero: `bool`, optional}}

        :Returns:

            {{Returns formula}}

        """
        standard_name = "ocean_s_coordinate_g1"

        computed_standard_name = cls._computed_standard_name(
            g, standard_name, coordinate_reference
        )

        # ----------------------------------------------------------------
        # Get the formula terms and their contruct keys
        # ----------------------------------------------------------------
        coordinate_conversion = coordinate_reference.coordinate_conversion

        s, s_key = cls._domain_ancillary_term(
            g,
            standard_name,
            coordinate_conversion,
            "s",
            default_to_zero,
            strict,
            True,
        )

        C, C_key = cls._domain_ancillary_term(
            g,
            standard_name,
            coordinate_conversion,
            "C",
            default_to_zero,
            strict,
            True,
        )

        eta, eta_key = cls._domain_ancillary_term(
            g,
            standard_name,
            coordinate_conversion,
            "eta",
            default_to_zero,
            strict,
            False,
        )

        depth, depth_key = cls._domain_ancillary_term(
            g,
            standard_name,
            coordinate_conversion,
            "depth",
            default_to_zero,
            strict,
            False,
        )

        depth_c, _ = cls._domain_ancillary_term(
            g,
            standard_name,
            coordinate_conversion,
            "depth_c",
            default_to_zero,
            strict,
            False,
        )

        cls._check_standard_name_consistency(
            strict, eta=(eta, eta_key), depth=(depth, depth_key)
        )

        # Make sure that eta and depth are consistent, which may require
        # eta to be transposed.
        eta, eta_axes = cls._conform_eta(g, eta, eta_key, depth, depth_key)

        # Get the axes of the non-parametric coordinates, putting the
        # vertical axis in postition -1 (the rightmost position).
        k_axis = cls._vertical_axis(g, [s_key, C_key])
        computed_axes = eta_axes + k_axis

        # Insert a size one dimension to allow broadcasting over the
        # vertical axis
        if k_axis:
            eta = eta.insert_dimension(-1)
            depth = depth.insert_dimension(-1)

        # ----------------------------------------------------------------
        # Compute the non-parametric coordinates
        # ----------------------------------------------------------------
        with bounds_combination_mode(
            "OR" if s.has_bounds() and C.has_bounds() else "AND"
        ):
            S = depth_c * s + (depth - depth_c) * C

            computed = S + eta * (1 + S / depth)

        return (
            standard_name,
            computed_standard_name,
            computed,
            computed_axes,
            k_axis,
        )

    @classmethod
    def ocean_s_coordinate_g2(
        cls, g, coordinate_reference, default_to_zero, strict
    ):
        """Compute non-parametric vertical coordinates from
        ocean_s_coordinate_g2 parametric coordinates.

        {{formula terms links}}

        .. note:: The vertical axis is the last (rightmost) dimension of
                  the returned computed non-parametric vertical
                  coordinates, if applicable.

        .. versionadded:: 3.8.0

        :Parameters:

            f: `Field`
                The parent field construct.

            coordinate_reference: `CoordinateReference`
                The coordinate reference construct of the parent field
                construct that defines the conversion formula.

            {{default_to_zero: `bool`, optional}}

        :Returns:

            {{Returns formula}}

        """
        standard_name = "ocean_s_coordinate_g2"

        computed_standard_name = cls._computed_standard_name(
            g, standard_name, coordinate_reference
        )

        # ----------------------------------------------------------------
        # Get the formula terms and their contruct keys
        # ----------------------------------------------------------------
        coordinate_conversion = coordinate_reference.coordinate_conversion

        s, s_key = cls._domain_ancillary_term(
            g,
            standard_name,
            coordinate_conversion,
            "s",
            default_to_zero,
            strict,
            True,
        )

        eta, eta_key = cls._domain_ancillary_term(
            g,
            standard_name,
            coordinate_conversion,
            "eta",
            default_to_zero,
            strict,
            False,
        )

        depth, depth_key = cls._domain_ancillary_term(
            g,
            standard_name,
            coordinate_conversion,
            "depth",
            default_to_zero,
            strict,
            False,
        )

        depth_c, _ = cls._domain_ancillary_term(
            g,
            standard_name,
            coordinate_conversion,
            "depth_c",
            default_to_zero,
            strict,
            False,
        )

        C, C_key = cls._domain_ancillary_term(
            g,
            standard_name,
            coordinate_conversion,
            "C",
            default_to_zero,
            strict,
            True,
        )

        cls._check_standard_name_consistency(
            strict, eta=(eta, eta_key), depth=(depth, depth_key)
        )

        # Make sure that eta and depth are consistent, which may require
        # eta to be transposed.
        eta, eta_axes = cls._conform_eta(g, eta, eta_key, depth, depth_key)

        # Get the axes of the non-parametric coordinates, putting the
        # vertical axis in postition -1 (the rightmost position).
        k_axis = cls._vertical_axis(g, [s_key, C_key])
        computed_axes = eta_axes + k_axis

        # Insert a size one dimension to allow broadcasting over the
        # vertical axis
        if k_axis:
            eta = eta.insert_dimension(-1)
            depth = depth.insert_dimension(-1)

        # ----------------------------------------------------------------
        # Compute the non-parametric coordinates
        # ----------------------------------------------------------------
        with bounds_combination_mode(
            "OR" if s.has_bounds() and C.has_bounds() else "AND"
        ):
            S = (depth_c * s + depth * C) / (depth + depth_c)

            computed = eta + (eta + depth) * S

        return (
            standard_name,
            computed_standard_name,
            computed,
            computed_axes,
            k_axis,
        )

    @classmethod
    def ocean_sigma_z_coordinate(
        cls, g, coordinate_reference, default_to_zero, strict
    ):
        """Compute non-parametric vertical coordinates from
        ocean_sigma_z_coordinate parametric coordinates.

        {{formula terms links}}

        .. note:: The vertical axis is the last (rightmost) dimension of
                  the returned computed non-parametric vertical
                  coordinates, if applicable.

        .. versionadded:: 3.8.0

        :Parameters:

            f: `Field`
                The parent field construct.

            coordinate_reference: `CoordinateReference`
                The coordinate reference construct of the parent field
                construct that defines the conversion formula.

            {{default_to_zero: `bool`, optional}}

        :Returns:

            {{Returns formula}}

        """
        standard_name = "ocean_sigma_z_coordinate"

        computed_standard_name = cls._computed_standard_name(
            g, standard_name, coordinate_reference
        )

        # ----------------------------------------------------------------
        # Get the formula terms and their contruct keys
        # ----------------------------------------------------------------
        coordinate_conversion = coordinate_reference.coordinate_conversion

        sigma, sigma_key = cls._domain_ancillary_term(
            g,
            standard_name,
            coordinate_conversion,
            "sigma",
            default_to_zero,
            strict,
            True,
        )

        eta, eta_key = cls._domain_ancillary_term(
            g,
            standard_name,
            coordinate_conversion,
            "eta",
            default_to_zero,
            strict,
            False,
        )

        depth, depth_key = cls._domain_ancillary_term(
            g,
            standard_name,
            coordinate_conversion,
            "depth",
            default_to_zero,
            strict,
            False,
        )

        depth_c, _ = cls._domain_ancillary_term(
            g,
            standard_name,
            coordinate_conversion,
            "depth_c",
            default_to_zero,
            strict,
            False,
        )

        nsigma, _ = cls._domain_ancillary_term(
            g,
            standard_name,
            coordinate_conversion,
            "nsigma",
            default_to_zero,
            strict,
            False,
        )

        zlev, zlev_key = cls._domain_ancillary_term(
            g,
            standard_name,
            coordinate_conversion,
            "zlev",
            default_to_zero,
            strict,
            True,
        )

        cls._check_standard_name_consistency(
            strict,
            zlev=(zlev, zlev_key),
            eta=(eta, eta_key),
            depth=(depth, depth_key),
        )

        # Make sure that eta and depth are consistent, which may require
        # eta to be transposed.
        eta, eta_axes = cls._conform_eta(g, eta, eta_key, depth, depth_key)

        # Get the axes of the non-parametric coordinates, putting the
        # vertical axis in postition -1 (the rightmost position).
        k_axis = cls._vertical_axis(g, [sigma_key, zlev_key])
        computed_axes = eta_axes + k_axis

        # Insert a size one dimension to allow broadcasting over the
        # vertical axis
        if k_axis:
            eta = eta.insert_dimension(-1)
            depth = depth.insert_dimension(-1)

        cls._check_index_term("nsigma", nsigma)

        # ----------------------------------------------------------------
        # Compute the non-parametric coordinates
        #
        # Note: This isn't overly efficient, because we do calculations
        #       for k>nsigma and then overwrite them.
        # ----------------------------------------------------------------
        with bounds_combination_mode(
            "OR" if zlev.has_bounds() and sigma.has_bounds() else "AND"
        ):
            computed = (
                eta + (depth.where(depth > depth_c, depth_c) + eta) * sigma
            )

            nsigma = int(nsigma.data)
            computed[..., nsigma:] = zlev[nsigma:]

        return (
            standard_name,
            computed_standard_name,
            computed,
            computed_axes,
            k_axis,
        )

    @classmethod
    def ocean_double_sigma_coordinate(
        cls, g, coordinate_reference, default_to_zero, strict
    ):
        """Compute non-parametric vertical coordinates from
        ocean_double_sigma_coordinate parametric coordinates.

        {{formula terms links}}

        .. note:: The vertical axis is the last (rightmost) dimension of
                  the returned computed non-parametric vertical
                  coordinates, if applicable.

        .. versionadded:: 3.8.0

        :Parameters:

            f: `Field`
                The parent field construct.

            coordinate_reference: `CoordinateReference`
                The coordinate reference construct of the parent field
                construct that defines the conversion formula.

            {{default_to_zero: `bool`, optional}}

        :Returns:

            {{Returns formula}}

        """
        standard_name = "ocean_double_sigma_coordinate"

        computed_standard_name = cls._computed_standard_name(
            g, standard_name, coordinate_reference
        )

        # ----------------------------------------------------------------
        # Get the formula terms and their contruct keys
        # ----------------------------------------------------------------
        coordinate_conversion = coordinate_reference.coordinate_conversion

        sigma, sigma_key = cls._domain_ancillary_term(
            g,
            standard_name,
            coordinate_conversion,
            "sigma",
            default_to_zero,
            strict,
            True,
        )

        depth, depth_key = cls._domain_ancillary_term(
            g,
            standard_name,
            coordinate_conversion,
            "depth",
            default_to_zero,
            strict,
            False,
        )

        z1, _ = cls._domain_ancillary_term(
            g,
            standard_name,
            coordinate_conversion,
            "z1",
            default_to_zero,
            strict,
            True,
        )

        z2, _ = cls._domain_ancillary_term(
            g,
            standard_name,
            coordinate_conversion,
            "z2",
            default_to_zero,
            strict,
            True,
        )

        a, _ = cls._domain_ancillary_term(
            g,
            standard_name,
            coordinate_conversion,
            "a",
            default_to_zero,
            strict,
            True,
        )

        href, _ = cls._domain_ancillary_term(
            g,
            standard_name,
            coordinate_conversion,
            "href",
            default_to_zero,
            strict,
            True,
        )

        k_c, _ = cls._domain_ancillary_term(
            g,
            standard_name,
            coordinate_conversion,
            "k_c",
            default_to_zero,
            strict,
            True,
        )

        # Get the axes of the non-parametric coordinates, putting the
        # vertical axis in postition -1 (the rightmost position).
        k_axis = cls._vertical_axis(g, [sigma_key])
        computed_axes = g.get_data_axes(depth_key) + k_axis

        # Insert a size one dimension to allow broadcasting over the
        # vertical axis
        if k_axis:
            depth = depth.insert_dimension(-1)

        cls._check_index_term("k_c", k_c)

        # ----------------------------------------------------------------
        # Compute the non-parametric coordinates
        #
        # Note: This isn't overly efficient, because we do calculations
        #       for k<=k_c and then overwrite them.
        # ----------------------------------------------------------------
        # Ensure that a, z1, z2, and href all have the same units as depth
        a = cls._conform_units("a", a, "depth", depth.Units)
        z1 = cls._conform_units("z1", z1, "depth", depth.Units)
        z2 = cls._conform_units("z2", z2, "depth", depth.Units)
        href = cls._conform_units("href", href, "depth", depth.Units)

        with bounds_combination_mode("OR" if sigma.has_bounds() else "AND"):
            f = (z1 + z2) * 0.5 + (
                0.5 * (z1 - z2) * (2 * a / (z1 - z2) * (depth - href)).tanh()
            )

            computed = f * sigma

            k_c1 = int(k_c.data) + 1

            computed[..., k_c1:] = f + (depth - f) * (sigma[k_c1:] - 1)

        return (
            standard_name,
            computed_standard_name,
            computed,
            computed_axes,
            k_axis,
        )

    @classmethod
    def formula(
        cls, f, coordinate_reference, default_to_zero=True, strict=True
    ):
        """Compute non-parametric vertical coordinates.

        {{formula terms links}}

        Dimensional vertical auxiliary coordinate values are computed from
        parametric vertical coordinate values (usually dimensionless) and
        associated domain ancillary constructs, as defined by the formula
        stored in a coordinate reference construct.

        .. versionadded:: 3.8.0

        :Parameters:

            f: `Field`
                The parent field construct.

            coordinate_reference: `CoordinateReference`
                A coordinate reference construct of the parent field
                construct.

            {{default_to_zero: `bool`, optional}}

            strict: `bool`
                If False then allow the computation to occur when

                * A domain ancillary construct has no standard name, but
                  the corresponding term has a standard name that is
                  prescribed

                * When the computed standard name can not be found by
                  inference from the standard names of the domain
                  ancillary constructs, nor from the
                  ``computed_standard_name`` parameter of the relevant
                  coordinate reference construct.

                By default an exception is raised in these cases.

                If a domain ancillary construct does have a standard name,
                but one that is inconsistent with any prescribed standard
                names, then an exception is raised regardless of the value
                of *strict*.

        :Returns:

            5-`tuple`
                * The standard name of the parametric coordinates.

                * The standard name of the computed non-parametric
                  coordinates. This may be `None` if a computed standard
                  name could not be found.

                * The computed coordinates in a `DomainAncillary`
                  construct.

                * A tuple of the domain axis construct keys for the
                  dimensions of the computed non-parametric coordinates.

                * A tuple containing the construct key of the vertical
                  domain axis. If the vertical axis does not appear in the
                  computed non-parametric coodinates then this an empty
                  tuple, instead.

                If the coordinate reference does not define a conversion
                from parametric coordinates to nonparametric coordinates
                then a `None` is returned for all of the tuple elements.

        """
        standard_name = (
            coordinate_reference.coordinate_conversion.get_parameter(
                "standard_name", None
            )
        )

        if standard_name is not None:
            logger.detail(
                f"standard_name: {standard_name!r}"
            )  # pragma: no cover

            if standard_name in cls.standard_names:
                # --------------------------------------------------------
                # Compute the non-parametric vertical coordinates
                # --------------------------------------------------------
                (
                    standard_name,
                    computed_standard_name,
                    computed,
                    computed_axes,
                    k_axis,
                ) = getattr(cls, standard_name)(
                    f, coordinate_reference, default_to_zero, strict
                )

                # --------------------------------------------------------
                # Move the vertical axis of the computed non-parametric
                # coordinates from its current position as the last
                # (rightmost) dimension, if applicable.
                # --------------------------------------------------------
                computed, computed_axes = cls._conform_computed(
                    f, computed, computed_axes, k_axis
                )

                return (
                    standard_name,
                    computed_standard_name,
                    computed,
                    computed_axes,
                    k_axis,
                )

        # Still here?
        return (None,) * 5



from functools import reduce
from operator import mul

import numpy

from ..functions import get_subspace, parse_indices
from . import abstract


class GatheredSubarray(abstract.CompressedSubarray):
    """An underlying gathered sub-array."""

    def __getitem__(self, indices):
        """x.__getitem__(indices) <==> x[indices]

        Returns a numpy array.

        """
        # The compressed array
        compressed_array = self.array

        # Initialize the full, uncompressed output array with missing
        # data everywhere
        uarray = numpy.ma.masked_all(self.shape, dtype=compressed_array.dtype)

        compression = self.compression
        compressed_dimension = compression["compressed_dimension"]
        compressed_axes = compression["compressed_axes"]
        compressed_part = compression["compressed_part"]
        list_array = compression["indices"]

        # Initialise the uncomprssed array
        n_compressed_axes = len(compressed_axes)

        uncompressed_shape = self.shape
        partial_uncompressed_shapes = [
            reduce(
                mul, [uncompressed_shape[i] for i in compressed_axes[j:]], 1
            )
            for j in range(1, n_compressed_axes)
        ]

        sample_indices = list(compressed_part)
        u_indices = [slice(None)] * self.ndim

        full = [slice(None)] * compressed_array.ndim

        zeros = [0] * n_compressed_axes
        for j, b in enumerate(list_array):
            # print('b=', b, end=", ")
            sample_indices[compressed_dimension] = slice(j, j + 1)

            # Note that it is important for indices a and b to be
            # integers (rather than the slices a:a+1 and b:b+1) so
            # that these dimensions are dropped from uarray[u_indices]
            u_indices[compressed_axes[0] : compressed_axes[-1] + 1] = zeros
            for i, z in zip(compressed_axes[:-1], partial_uncompressed_shapes):
                if b >= z:
                    (a, b) = divmod(b, z)
                    u_indices[i] = a
            # --- End: for
            u_indices[compressed_axes[-1]] = b

            compressed = compressed_array[tuple(sample_indices)].array
            sample_indices2 = full[:]
            sample_indices2[compressed_dimension] = 0
            compressed = compressed[tuple(sample_indices2)]

            uarray[tuple(u_indices)] = compressed
        # --- End: for

        if indices is Ellipsis:
            return uarray
        else:
            indices = parse_indices(self.shape, indices)
            return get_subspace(uarray, indices)


# --- End: class



import cfdm


class Array(cfdm.Array):
    """Abstract base class for a container of an underlying array.

    The form of the array is defined by the initialization parameters
    of a subclass.

    .. versionadded:: 3.0.0

    """

    def __repr__(self):
        """Called by the `repr` built-in function.

        x.__repr__() <==> repr(x)

        .. versionadded:: 3.0.0

        """
        return super().__repr__().replace("<", "<CF ", 1)



import datetime
import faulthandler
import unittest

faulthandler.enable()  # to debug seg faults and timeouts

import cf


class ListTest(unittest.TestCase):
    gathered = "gathered.nc"

    def setUp(self):
        # Disable log messages to silence expected warnings
        cf.log_level("DISABLE")
        # Note: to enable all messages for given methods, lines or
        # calls (those without a 'verbose' option to do the same)
        # e.g. to debug them, wrap them (for methods, start-to-end
        # internally) as follows:
        #
        # cf.LOG_LEVEL('DEBUG')
        # < ... test code ... >
        # cf.log_level('DISABLE')

    def test_List__repr__str__dump(self):
        f = cf.read(self.gathered)[0]

        lst = f.data.get_list()

        repr(lst)
        str(lst)
        self.assertIsInstance(lst.dump(display=False), str)


if __name__ == "__main__":
    print("Run date:", datetime.datetime.now())
    cf.environment()
    print()
    unittest.main(verbosity=2)



import atexit
import logging
from copy import deepcopy
from functools import reduce
from itertools import product as itertools_product
from operator import mul
from os import close, remove, rmdir
from os.path import isfile
from sys import getrefcount
from tempfile import mkstemp

from numpy import array as numpy_array
from numpy import bool_ as numpy_bool_
from numpy import dtype as numpy_dtype
from numpy import ndarray as numpy_ndarray
from numpy import number as numpy_number
from numpy import transpose as numpy_transpose
from numpy import vectorize as numpy_vectorize
from numpy.ma import MaskedArray as numpy_ma_MaskedArray
from numpy.ma import is_masked as numpy_ma_is_masked
from numpy.ma import isMA as numpy_ma_isMA
from numpy.ma import masked_all as numpy_ma_masked_all
from numpy.ma import nomask as numpy_ma_nomask

from ..functions import fm_threshold as cf_fm_threshold
from ..functions import free_memory, get_subspace
from ..functions import inspect as cf_inspect
from ..units import Units
from .abstract import FileArray

# from .filearray import  (_TempFileArray #, SharedMemoryArray,
#                          _shared_memory_array,FileArray)
from .cachedarray import CachedArray

# from cfunits import Units


logger = logging.getLogger(__name__)

_Units_conform = Units.conform

_dtype_object = numpy_dtype(object)

# --------------------------------------------------------------------
# Dictionary of partitions' temporary files containing the full path
# of the directory containing tuples of the temporary file and its
# lock files, the full path of the lock file on this PE and a set of
# the full path of the lock files on other PEs.
#
# For example:
# >>> _temporary_files
# {'/tmp/cf_array_7hoQk5/cf_array_RmC7NJ.npy': ('/tmp/cf_array_7hoQk5',
#  '/tmp/cf_array_7hoQk5/cf_array_RmC7NJ.npy_i043rk',
#  set('/tmp/cf_array_7hoQk5/cf_array_RmC7NJ.npy_cvUcib')),
#  '/tmp/cf_array_CrbJVf/cf_array_INZQQc.npy': ('/tmp/cf_array_CrbJVf',
#  '/tmp/cf_array_CrbJVf/cf_array_INZQQc.npy_n4BKBF',
#  set('/tmp/cf_array_CrbJVf/cf_array_INZQQc.npy_xQ5LVo'))}
# --------------------------------------------------------------------
_temporary_files = {}


def _lock_files_present(lock_files):
    lock_files_present = False
    for filename in lock_files:
        if isfile(filename):
            lock_files_present = True
            break
    # --- End: for

    return lock_files_present


def _remove_temporary_files(filename=None):
    """Remove temporary partition files from disk.

    The removed files' names are deleted from the _temporary_files
    set.

    The intended use is to delete individual files as part of the
    garbage collection process and to delete all files when python
    exits.

    This is quite brutal and may break partitions if used unwisely. It
    is not recommended to be used as a general tidy-up function.

    .. seealso:: `__del__`

    :Parameters:

        filename: `str`, optional
            The name of file to remove. The file name must be in the
            _temporary_files set. By default all files named in the
            _temporary_files set are removed.

    :Returns:

        `None`

    **Examples:**

    >>> _remove_temporary_files()

    >>> _temporary_files
    {'/tmp/cf_array_7hoQk5/cf_array_RmC7NJ.npy': ('/tmp/cf_array_7hoQk5',
     '/tmp/cf_array_7hoQk5/cf_array_RmC7NJ.npy_i043rk',
     set('/tmp/cf_array_7hoQk5/cf_array_RmC7NJ.npy_cvUcib')),
     '/tmp/cf_array_CrbJVf/cf_array_INZQQc.npy': ('/tmp/cf_array_CrbJVf',
     '/tmp/cf_array_CrbJVf/cf_array_INZQQc.npy_n4BKBF',
     set('/tmp/cf_array_CrbJVf/cf_array_INZQQc.npy_xQ5LVo'))}
    >>> _remove_temporary_files('/tmp/cf_array_7hoQk5/cf_array_RmC7NJ.npy')
    >>> _temporary_files
    {'/tmp/cf_array_CrbJVf/cf_array_INZQQc.npy': ('/tmp/cf_array_CrbJVf',
     '/tmp/cf_array_CrbJVf/cf_array_INZQQc.npy_n4BKBF',
     set('/tmp/cf_array_CrbJVf/cf_array_INZQQc.npy_xQ5LVo'))}
    >>> _remove_temporary_files()
    >>> _temporary_files
    {}

    """
    if filename is not None:
        if filename in _temporary_files:
            # If this condition is not met then probably
            # _remove_temporary_files() has already been run at
            # exit
            dirname, _lock_file, _other_lock_files = _temporary_files[filename]
            try:
                remove(_lock_file)
            except OSError:
                pass

            # Only remove the temporary file if it is not being
            # used by any other ranks
            if not _lock_files_present(_other_lock_files):
                # Remove the given temporary file
                try:
                    remove(filename)
                    rmdir(dirname)
                except OSError:
                    pass
                del _temporary_files[filename]
        # --- End: if

        return

    # Still here? Then remove all temporary files and lock files
    for filename in _temporary_files:
        try:
            remove(filename)
        except OSError:
            pass
        dirname, _lock_file, _other_lock_files = _temporary_files[filename]
        try:
            remove(_lock_file)
        except OSError:
            pass
        for lock_file in _other_lock_files:
            try:
                remove(lock_file)
            except OSError:
                pass
        # --- End: for

        try:
            rmdir(dirname)
        except OSError:
            pass
    # --- End: for

    _temporary_files.clear()


# --------------------------------------------------------------------
# Instruction to remove all of the temporary files from all partition
# arrays at exit.
# --------------------------------------------------------------------
atexit.register(_remove_temporary_files)

# --------------------------------------------------------------------
# Create a deep copy function for numpy arrays which contain object
# types
# --------------------------------------------------------------------
_copy = numpy_vectorize(deepcopy, otypes=[object])


class Partition:
    """A partition of a master data array.

    The partition spans all or part of exactly one subarray of the
    master data array

    """

    # Counters for the number of partitions pointing to each open
    # file.  A file is only closed when the count reaches zero for
    # that file.  The key is the absolute path to the file. The
    # corresponding value is the counter.
    file_counter = {}

    def __init__(
        self,
        subarray=None,
        flip=None,
        location=None,
        shape=None,
        Units=None,
        part=None,
        axes=None,
        fill=None,
    ):
        """**Initialization**

        :Parameters:

            subarray: numpy array-like, optional
                The subarray for the partition. Must be a numpy array or
                any array storing object with a similar interface. DO NOT
                UPDATE INPLACE.

            location: `list`, optional
                The location of the partition's data array in the master
                array. DO NOT UPDATE INPLACE.

            axes: `list`, optional
                The identities of the axes of the partition's subarray. If
                the partition's subarray a scalar array then it is an
                empty list. DO NOT UPDATE INPLACE.

            part: `list`, optional
                The part of the partition's subarray which comprises its
                data array. If the partition's data array is to the whole
                subarray then *part* may be an empty list. DO NOT UPDATE
                INPLACE.

            shape: `list`, optional
                The shape of the partition's data array as a subspace of
                the master array. If the master array is a scalar array
                then *shape* is an empty list. By default the shape is
                inferred from *location*. DO NOT UPDATE INPLACE.

            Units: `Units`, optional
                The units of the partition's subarray. DO NOT UPDATE
                INPLACE.

        **Examples:**

        >>> p = Partition(subarray   = numpy.arange(20).reshape(2,5,1),
        ...               location   = [(0, 6), (1, 3), (4, 5)],
        ...               axes       = ['dim1', 'dim0', 'dim2'],
        ...               part       = [],
        ...               Units      = cf.Units('K'))

        >>> p = Partition(subarray   = numpy.arange(20).reshape(2,5,1),
        ...               location   = [(0, 6), (1, 3), (4, 5)],
        ...               axes       = ['dim1', 'dim0', 'dim2'],
        ...               shape      = [5, 2, 1],
        ...               part       = [slice(None, None, -1), [0,1,3,4],
        ...                             slice(None)],
        ...               Units      = cf.Units('K'))

        >>> p = Partition(subarray   = numpy.array(4),
        ...               location   = [(4, 5)],
        ...               axes       = ['dim1'],
        ...               part       = [],
        ...               Units      = cf.Units('K'))

        """

        self._subarray = None

        self.axes = axes  # DO NOT UPDATE INPLACE
        self.flip = flip  # DO NOT UPDATE INPLACE
        self.part = part  # DO NOT UPDATE INPLACE
        self.location = location  # DO NOT UPDATE INPLACE
        self.shape = shape  # DO NOT UPDATE INPLACE
        self.Units = Units  # DO NOT UPDATE INPLACE
        self.subarray = subarray  # DO NOT UPDATE INPLACE
        self.fill = fill  # DO NOT UPDATE INPLACE

        if shape is None and location is not None:
            self.shape = [i[1] - i[0] for i in location]

        self._write_to_disk = None

    def __deepcopy__(self, memo):
        """Used if copy.deepcopy is called on the variable."""
        return self.copy()

    def __del__(self):
        """Called when the partition's reference count reaches zero.

        If the partition contains a temporary file which is not referenced
        by any other partition then the temporary file is removed from
        disk.

        If the partition contains a non-temporary file which is not
        referenced by any other partition then the file is closed.

        """
        #        subarray = getattr(self, '_subarray', None)
        subarray = self._subarray

        # If the subarray is unique it will have 2 references to
        # it plus 1 within this method, making 3. If it has more
        # than 3 references to it then it is not unique.
        if getrefcount is not None:
            self._decrement_file_counter()
            if subarray is None or getrefcount(subarray) > 3:
                return
        else:
            # getrefcount has itself been deleted or is in the process
            # of being torn down
            return

        _partition_file = getattr(subarray, "_partition_file", None)
        if _partition_file is not None:
            # This partition contains a temporary file which is not
            # referenced by any other partition on this process, so if
            # there are no lock files present remove the file from
            # disk.
            _remove_temporary_files(_partition_file)

        else:
            try:
                if FileArray is not None and isinstance(subarray, FileArray):
                    try:
                        filename = subarray.get_filename()
                    except Exception:
                        filename = None

                    if self.file_counter.get(filename, 999) <= 0:
                        # This partition contains a non-temporary file
                        # which is not referenced by any other
                        # partitions, so close the file.
                        subarray.close()
            except Exception:
                # If we're here then it is likely that FileArray has been
                # torn down, so just do nothing.
                pass
        # --- End: if

    #    def __getstate__(self):
    #        '''
    #
    #    Called when pickling.
    #
    #    :Parameters:
    #
    #        None
    #
    #    :Returns:
    #
    #        `dict`
    #            A dictionary of the instance's attributes
    #
    #    **Examples:**
    #
    #        '''
    #        return dict([(attr, getattr(self, attr))
    #                     for attr in self.__slots__ if hasattr(self, attr)])
    #
    #    def __setstate__(self, odict):
    #        '''
    #
    #    Called when unpickling.
    #
    #    :Parameters:
    #
    #        odict : dict
    #            The output from the instance's `__getstate__` method.
    #
    #    :Returns:
    #
    #        None
    #
    #        '''
    #        for attr, value in odict.items():
    #            setattr(self, attr, value)

    def __str__(self):
        """x.__str__() <==> str(x)"""
        return "%s: %s" % (self.__class__.__name__, self.__dict__)

    def _add_to_file_counter(self, i):
        """Add i to the count of subarrays referencing the file of this
        partition's subarray.

        Only do this if self._subarray is an instance of FileArray, but
        not a temporary FileArray.

        :Parameters:

            i: `int`

        :Returns:

            `None`

        """
        #        subarray = getattr(self, '_subarray', None)
        subarray = self._subarray

        if subarray is None:
            return

        try:
            if isinstance(subarray, FileArray) and not isinstance(
                subarray, CachedArray
            ):
                try:
                    filename = subarray.get_filename()
                except Exception:
                    filename = None

                if filename is None:
                    return

                file_counter = self.file_counter
                #                count = file_counter.get(filename, 0)
                #                file_counter[filename] = count + i
                #                if file_counter[filename] <= 0:
                count = file_counter.get(filename, 0) + i
                if count <= 0:
                    # Remove the file from the dictionary if its count has
                    # dropped to zero
                    file_counter.pop(filename, None)
                else:
                    file_counter[filename] = count
        except Exception:
            # If we're here then it is likely that FileArray has been
            # torn down, so just do nothing.
            pass

    def _increment_file_counter(self):
        """Add 1 to the Partition.file_counter if self._subarray is an
        instance of FileArray and not a temporary FileArray.

        :Returns:

            `None`

        """
        self._add_to_file_counter(1)

    def _decrement_file_counter(self):
        """Subtract 1 from the Partition.file_counter if self._subarray
        is an instance of FileArray and not a temporary FileArray.

        :Returns:

            `None`

        """
        self._add_to_file_counter(-1)

    def _configure_auxiliary_mask(self, auxiliary_mask):
        """Add the auxiliary mask to the config dictionary.

        Assumes that ``self.config`` already exists.

        :Parameters:

            auxiliary_mask: `list` of `Data`

        :Returns:

            `None`

        **:Examples:**

        >>> p._configure_auxiliary_mask([mask_component1, mask_component2])

        """
        indices = self.indices

        new = [
            mask[
                tuple(
                    [
                        (slice(None) if n == 1 else index)
                        for n, index in zip(mask.shape, indices)
                    ]
                )
            ]
            for mask in auxiliary_mask
        ]

        #        # If the partition is to be parallelised then get rid of mask
        #        # components which are all False so the mask component does
        #        # not get copied to the child process
        #        if not config['serial']:
        #            new = [mask for mask in new if not mask.any()]

        self.config["auxiliary_mask"] = new

    # ----------------------------------------------------------------
    # Attributes
    # ----------------------------------------------------------------
    @property
    def indices(self):
        """The indices of the master array which correspond to this
        partition's data array.

        :Returns:

            `tuple`
                A tuple of slice objects or, if the master data array is a
                scalar array, an empty tuple.

        **Examples:**

        >>> p.location
        [(0, 5), (2, 9)]
        >>> p.indices
        (slice(0, 5), slice(2, 9))

        >>> p.location
        [()]
        >>> p.indices
        ()

        """
        return tuple([slice(*r) for r in self.location])

    @property
    def in_memory(self):
        """True if and only if the partition's subarray is in memory as
        opposed to on disk.

        **Examples:**

        >>> p.in_memory
        False

        """
        return hasattr(self._subarray, "__array_interface__")

    #    @property
    #    def in_shared_memory(self):
    #        '''
    #
    #    True if and only if the partition's subarray is a shared memory array.
    #
    #    .. seealso:: `array`, `in_memory`, `on_disk`, `to_shared_memory`
    #
    #    **Examples:**
    #
    #    >>> p.in_shared_memory
    #    True
    #    >>> p.to_disk()
    #    >>> p.in_shared_memory
    #    False
    #
    #        '''
    #        return isinstance(self._subarray, SharedMemoryArray)

    @property
    def in_cached_file(self):
        """True if and only if the partition's subarray is on disk in a
        temporary file.

        .. seealso:: `array`, `in_memory`, `in_shared_memory`, `on_disk`,
                     `to_disk`

        **Examples:**

        >>> p.in_cached_file
        False

        """
        return isinstance(self._subarray, CachedArray)

    @property
    def on_disk(self):
        """True if and only if the partition's subarray is on disk as
        opposed to in memory.

        **Examples:**

        >>> p.on_disk
        True
        >>> p.to_memory()
        >>> p.on_disk
        False

        """
        return isinstance(self._subarray, FileArray)

    @property
    def in_file(self):
        """True if and only if the partition's subarray is on disk as
        opposed to in memory.

        **Examples:**

        >>> p.on_disk
        True
        >>> p.to_disk()
        >>> p.on_disk
        False

        """
        return self.on_disk and not self.in_cached_file

    @property
    def dtype(self):
        """The data type of the master array."""
        return self.config["dtype"]

    @property
    def isscalar(self):
        """True if and only if the partition's data array is a scalar
        array.

        **Examples:**

        >>> p.axes
        []
        >>> p.isscalar
        True

        >>> p.axes
        ['dim2']
        >>> p.isscalar
        False

        """
        return not self.axes

    @property
    def nbytes(self):
        """The size in bytes of the subarray.

        The size takes into account the datatype and assumes that there
        is a boolean mask, unless it can be ascertained that there isn't
        one.

        """
        dtype = self.config["dtype"]
        if dtype is None:
            return None

        size = reduce(mul, self.shape, 1)
        nbytes = size * dtype.itemsize

        if getattr(self, "masked", True):
            nbytes += size

        return nbytes

    @property
    def ndim(self):
        """Number of array dimensions."""
        return len(self.shape)

    @property
    def size(self):
        """Number of elements in the partition's data array (not its
        subarray).

        **Examples:**

        >>> p.shape
        (73, 48)
        >>> p.size
        3504

        """
        return reduce(mul, self.shape, 1)

    @property
    def subarray(self):
        """The partition's subarray of data."""
        return self._subarray

    @subarray.setter
    def subarray(self, value):
        self._decrement_file_counter()
        self._subarray = value
        self._increment_file_counter()
        self._in_place_changes = False

    @subarray.deleter
    def subarray(self):
        self._decrement_file_counter()
        self._subarray = None
        self._in_place_changes = True

    #    @property
    #    def subarray_in_external_file(self):
    #        '''
    #
    #    True if and only if the partition's subarray is in an external file.
    #
    #    **Examples:**
    #
    #    >>> p.subarray_in_external_file
    #    False
    #
    #        '''
    #
    #        return not (self.in_memory or isinstance(self.subarray, FileArray))

    def change_axis_names(self, axis_map):
        """Change the axis names.

        The axis names are arbitrary, so mapping them to another arbitrary
        collection does not change the data array values, units, nor axis
        order.

        :Parameters:

            axis_map: `dict`

        :Returns:

            `None`

        **Examples:**

        >>> p.axes
        ['dim0', 'dim1']
        >>> p._change_axis_names({'dim0': 'dim2', 'dim1': 'dim0'})
        >>> p.axes
        ['dim2', 'dim0']

        >>> p.axes
        ['dim0', 'dim1']
        >>> p._change_axis_names({'dim0': 'dim1'})
        >>> p.axes
        ['dim1', 'dim2']

        """
        axes = self.axes

        # Partition axes
        self.axes = [axis_map[axis] for axis in axes]

        # Flipped axes
        flip = self.flip
        if flip:
            self.flip = [axis_map[axis] for axis in flip]

    def close(self, **kwargs):
        """Close the partition after it has been conformed.

        The partition should usually be closed after its `array` method
        has been called to prevent memory leaks.

        Closing the partition does one of the following, depending on the
        values of the partition's `!_original` attribute and on the
        *keep_in_memory* argument:

        * Nothing.

        * Stores the partition's data array in a temporary file.

        * Reverts the entire partition to a previous state.

        :Parameters:

            to_disk: `bool`, optional
                If True then revert to file pointer or write to disk
                regardless. Ignored if False

            in_memory: `bool`, optional
                If True then keep in memory, if possible,
                regardless. Ignored if False

        :Returns:

            None

        **Examples:**

        >>> p.array(...)
        >>> p.close()

        """
        config = getattr(self, "config", None)

        if config is None:
            return

        if kwargs:
            config.update(kwargs)

        original = getattr(self, "_original", None)
        logger.debug("Partition.close: original = {}".format(original))

        if not original:
            originally_on_disk = False
            original_subarray = None
        else:
            originally_on_disk = not original.in_memory
            original_subarray = original._subarray

        config = self.config
        logger.debug(" config = {}".format(config))

        if config["serial"]:
            # --------------------------------------------------------
            # SERIAL
            # --------------------------------------------------------
            logger.debug("  serial")

            if config["readonly"]:
                logger.debug("   readonly=True")

                if originally_on_disk:
                    logger.debug("    subarray originally on disk")

                    if config.get("to_disk", False):
                        # 1.1.1.1 The original subarray was on disk,
                        #         we don't want to keep the current
                        #         subarray in memory, and we are happy
                        #         to discard any changes that may have
                        #         been made to the subarray.
                        logger.debug("    1.1.1.1 revert")
                        self.revert()
                    elif free_memory() <= cf_fm_threshold():
                        # 1.1.1.2 The original subarray was on disk,
                        #         we are happy to keep the current
                        #         subarray in memory, but there is not
                        #         enough free memory to do so.
                        logger.debug(
                            "    1.1.1.2 revert ({} <= {})".format(
                                free_memory(), cf_fm_threshold()
                            )
                        )
                        self.revert()
                    else:
                        # 1.1.1.3 The original subarray was on disk
                        #         and there is enough memory to keep
                        #         the current subarray in memory
                        if config["unique_subarray"] and isinstance(
                            original_subarray, CachedArray
                        ):
                            # The original subarray was a temporary
                            # file which is not referenced by any
                            # other partitions
                            _remove_temporary_files(
                                original_subarray._partition_file
                            )

                        del self.masked
                        logger.debug(
                            "    1.1.1.3 del masked ({} > {})".format(
                                free_memory(), cf_fm_threshold()
                            )
                        )

                else:
                    logger.debug("   subarray originally in memory")
                    if config.get("to_disk", False):
                        # 1.1.2.1 Original subarray was in memory and
                        #         we don't want to keep the current
                        #         subarray in memory
                        logger.debug("    1.1.2.1 to_disk")
                        self.to_disk(reopen=False)
                    elif free_memory() <= cf_fm_threshold():
                        # 1.1.2.2 Original subarray was in memory and
                        #         unique but there is not enough
                        #         memory to keep the current subarray
                        logger.debug("    1.1.2.2 to_disk")
                        self.to_disk(reopen=False)
                    else:
                        # 1.1.2.3 Original subarray was in memory and
                        #         unique and there is enough memory to
                        #         keep the current subarray in memory
                        logger.debug("    1.1.2.3 pass")
                        pass
            else:
                # config['readonly'] is False
                if originally_on_disk:
                    if config.get("to_disk", False):
                        # 1.2.1.1 Original subarray was on disk and
                        #         there and we don't want to keep the
                        #         array
                        if config["unique_subarray"] and isinstance(
                            original_subarray, CachedArray
                        ):
                            # Original subarray was a temporary file
                            # on disk which is not referenced by any
                            # other partitions
                            _remove_temporary_files(
                                original_subarray._partition_file
                            )

                        logger.debug("    1.2.1.1 to_disk")
                        self.to_disk(reopen=False)
                    elif free_memory() <= cf_fm_threshold():
                        # 1.2.1.2 Original subarray was on disk but
                        #         there is not enough memory to keep
                        #         it
                        if config["unique_subarray"] and isinstance(
                            original_subarray, CachedArray
                        ):
                            # Original subarray was a temporary file
                            # on disk which is not referenced by any
                            # other partitions
                            _remove_temporary_files(
                                original_subarray._partition_file
                            )

                        logger.debug("    1.2.1.2 to_disk")
                        self.to_disk(reopen=False)
                    else:
                        # 1.2.1.3 Original subarray was on disk and
                        #         there is enough memory to keep it
                        logger.debug("    1.2.1.3 pass")
                        del self.masked
                else:
                    if config.get("to_disk", False):
                        # 1.2.2.1 Original subarray was in memory but
                        #         we don't want to keep it
                        logger.debug("    1.2.2.1 to_disk")
                        self.to_disk(reopen=False)
                    elif free_memory() <= cf_fm_threshold():
                        # 1.2.2.2 Original subarray was an in memory
                        #         but there is not enough memory to
                        #         keep it
                        logger.debug("    1.2.2.2 to_disk")
                        self.to_disk(reopen=False)
                    else:
                        # 1.2.2.3 Original subarray was in memory and
                        #         there is enough memory to keep it
                        logger.debug("    1.2.2.3 del masked")
                        del self.masked
        else:
            logger.debug("Partition.close: parallel")
            # --------------------------------------------------------
            # PARALLEL
            # --------------------------------------------------------
            pass

        #        if hasattr(self, '_original'):
        #            del self._original

        #        print(hasattr(self, 'config')),
        try:
            del self.config
        except AttributeError:
            pass

    def copy(self):
        """Return a deep copy.

        ``p.copy()`` is equivalent to ``copy.deepcopy(p)``.

        :Returns:

                A deep copy.

        **Examples:**

        >>> q = p.copy()

        """
        new = Partition.__new__(Partition)
        new.__dict__ = self.__dict__.copy()

        self._increment_file_counter()

        return new

    @property
    def array(self):
        """Returns the partition's data array.

        After a partition has been conformed, the partition must be closed
        (with the `close` method) before another partition is conformed,
        otherwise a memory leak could occur. For example:

        >>> for partition in partition_array.flat:
        ...    # Open the partition
        ...    partition.open(config)
        ...
        ...    # Get the data array as a numpy array
        ...    array = partition.array
        ...
        ...    # < Some code to operate on the array >
        ...
        ...    # Close the partition
        ...    partition.close()
        ...
        ...    # Now move on to conform the next partition
        ...
        >>>

        """
        config = self.config

        unique_array = config["unique_subarray"]

        p_axes = self.axes
        p_flip = self.flip
        p_part = self.part
        p_units = self.Units
        p_shape = self.shape
        p_location = self.location
        subarray = self._subarray

        len_p_axes = len(p_axes)

        if not self.in_memory:
            # --------------------------------------------------------
            # The subarray is not in memory.
            #
            # It could be in a file on disk or implied by a FileArray
            # object, etc.
            # --------------------------------------------------------
            self._original = self.copy()

            unique_array = True
            update = True
            copy = False

            if not p_part:
                indices = Ellipsis
            else:
                indices = tuple(p_part)

            # Read from a file into a numpy array
            p_data = subarray[indices]

            # We've just copied p_data from disk, so in place changes
            # are not possible
            in_place_changes = False
        else:
            # --------------------------------------------------------
            # The subarray is in memory
            # --------------------------------------------------------
            update = config["update"]

            if p_part:
                p_data = get_subspace(subarray, p_part)
            elif not unique_array:
                p_data = subarray.view()
            else:
                p_data = subarray

            copy = config["extra_memory"]

            # In place changes to p_data might be possible if we're not
            # copying the data
            in_place_changes = not copy

        if not p_data.ndim and isinstance(p_data, (numpy_number, numpy_bool_)):
            # --------------------------------------------------------
            # p_data is a numpy number (like numpy.int64) which does
            # not support assignment, so convert it to a numpy array.
            # --------------------------------------------------------
            p_data = numpy_array(p_data)
            # We've just copied p_data, so in place changes are
            # not possible
            copy = False
            in_place_changes = False

        masked = numpy_ma_isMA(p_data)
        if masked:
            # The p_data is a masked array
            if p_data.mask is numpy_ma_nomask or not numpy_ma_is_masked(
                p_data
            ):
                # There are no missing data points so recast as an
                # unmasked numpy array
                p_data = p_data.data
                masked = False
        # --- End: if

        if masked:
            # Set the hardness of the mask
            if config["hardmask"]:
                p_data.harden_mask()
            else:
                p_data.soften_mask()
        # --- End: if

        self.masked = masked

        # ------------------------------------------------------------
        # Make sure that the data array has the correct units. This
        # process will deep copy the data array if required (e.g. if
        # another partition is referencing this numpy array), even if
        # the units are already correct.
        # ------------------------------------------------------------
        func = config.get("func")
        units = config["units"]
        if func is None:
            if not p_units.equals(units) and bool(p_units) is bool(units):
                func = Units.conform

        if func is not None:
            inplace = not copy
            p_data = func(p_data, p_units, units, inplace)
            p_units = units

            if not inplace:
                # We've just copied p_data, so in place changes are
                # not possible
                copy = False
                in_place_changes = False
        # --- End: if

        flip = config.get("flip", None)
        if flip or p_flip:
            flip_axes = set(p_flip).symmetric_difference(flip)
        else:
            flip_axes = None

        axes = config["axes"]

        if p_data.size > 1:
            # --------------------------------------------------------
            # Flip axes
            # --------------------------------------------------------
            if flip_axes:
                indices = [
                    (
                        slice(None, None, -1)
                        if axis in flip_axes
                        else slice(None)
                    )
                    for axis in p_axes
                ]
                p_data = p_data[tuple(indices)]

            # --------------------------------------------------------
            # Transpose axes
            # --------------------------------------------------------
            if p_axes != axes:
                iaxes = [p_axes.index(axis) for axis in axes if axis in p_axes]

                if len_p_axes > len(iaxes):
                    for i in range(len_p_axes):
                        if i not in iaxes:
                            # iaxes.append(i)
                            iaxes.insert(i, i)
                # --- End: if

                p_data = numpy_transpose(p_data, iaxes)
        # --- End: if

        # ------------------------------------------------------------
        # Remove excessive/insert missing size 1 axes
        # ------------------------------------------------------------
        if p_shape != p_data.shape:
            # if len_p_axes != len(p_shape):
            p_data = p_data.reshape(p_shape)

        # ------------------------------------------------------------
        # Apply the auxiliary mask
        # ------------------------------------------------------------
        auxiliary_mask = config["auxiliary_mask"]
        if auxiliary_mask:
            for mask in auxiliary_mask:
                if mask.any():
                    if not masked:
                        p_data = p_data.view(numpy_ma_MaskedArray)
                        masked = True

                    p_data.mask = (mask | p_data.mask).array
            # --- End: for

            self.masked = True

        # ------------------------------------------------------------
        # Convert the array's data type
        # ------------------------------------------------------------
        p_dtype = p_data.dtype
        dtype = config.get("dtype", None)
        if dtype is not None and dtype != p_dtype:
            try:
                p_data = p_data.astype(dtype)  # Note: returns a copy
            except ValueError:
                raise ValueError(
                    "Can't recast partition array from {} to {}".format(
                        p_dtype.name, dtype.name
                    )
                )
            else:
                # We've just copied p_data, so in place changes are
                # not possible
                copy = False
                in_place_changes = False
        # --- End: if

        # ------------------------------------------------------------
        # Copy the array
        # -----------------------------------------------------------
        if copy:
            if p_dtype.char != "O":
                if not masked or p_data.ndim > 0:
                    p_data = p_data.copy()
                else:
                    # This is because numpy.ma.copy doesn't work for
                    # scalar arrays (at the moment, at least)
                    p_data = numpy_ma_masked_all((), p_data.dtype)

                # We've just copied p_data, so in place changes are
                # not possible
                in_place_changes = False
            else:
                # whilst netCDF4.netcdftime.datetime is mucking bout,
                # don't copy!!!!
                # p_data = _copy(p_data)
                pass
        # --- End: if

        # ------------------------------------------------------------
        # Update the partition
        # ------------------------------------------------------------
        if update:
            self.subarray = p_data  # ?? DCH CHECK
            self.Units = p_units
            self.part = []
            self.axes = axes
            self.flip = flip
            self.flatten = []
            self.shape = p_shape
            self.location = p_location

            self._in_place_changes = in_place_changes

        # ------------------------------------------------------------
        # Return the numpy array
        # ------------------------------------------------------------
        return p_data

    @property
    def isdt(self):
        """True if the subarray contains date-time objects.

        **Examples:**

        >>> p.Units.isreftime
        True
        >>> p.subarray.dtype == numpy.dtype(object)
        True
        >>> p.isdt
        True

        """
        return self.Units.isreftime and self._subarray.dtype == _dtype_object

    def file_close(self):
        """Close the file containing the subarray, if there is one.

        :Returns:

            `None`

        **Examples:**

        >>> p.file_close()

        """
        if self.on_disk:
            self._subarray.close()

    #    def flat(self):
    #        '''
    #
    #    Return an iterator that yields the partition itself.
    #
    #    This is provided as a convienience to make it easier to iterate
    #    through a partition matrix.
    #
    #    :Returns:
    #
    #        generator
    #            An iterator that yields the partition itself.
    #
    #    **Examples:**
    #
    #    >>> type(p.flat())
    #    <generator object flat at 0x519a0a0>
    #    >>> for q in p.flat():
    #    ...    print(q is p)
    #    True
    #
    #    '''
    #            yield self
    #
    #
    #    def ndindex(self):
    #        '''
    #
    #    Return an iterator over the N-dimensional indices of the partition's
    #    data array.
    #
    #    At each iteration a tuple of indices is returned, the last dimension
    #    is iterated over first.
    #
    #    :Returns:
    #
    #        generator
    #            An iterator over indices of the partition's data array.
    #
    #    **Examples:**
    #
    #    >>> p.shape
    #    [2, 1, 3]
    #    >>> for index in p.ndindex():
    #    ...     print(index)
    #    ...
    #    (0, 0, 0)
    #    (0, 0, 1)
    #    (0, 0, 2)
    #    (1, 0, 0)
    #    (1, 0, 1)
    #    (1, 0, 2)
    #
    #    >>> p.shape
    #    []
    #    >>> for index in p.ndindex():
    #    ...     print(index)
    #    ...
    #    ()
    #
    #        '''
    #        return itertools_product(*[range(0, r) for r in self.shape])

    def inspect(self):
        """Inspect the object for debugging.

        .. seealso:: `cf.inspect`

        :Returns:

            None

        **Examples:**

        >>> f.inspect()

        """
        print(cf_inspect(self))

    def master_ndindex(self):  # itermaster_indices(self):
        """Return an iterator over indices of the master array which are
        spanned by the data array.

        :Returns:

            generator
                An iterator over indices of the master array which are
                spanned by the data array.

        **Examples:**

        >>> p.location
        [(3, 5), (0, 1), (0, 3)]
        >>> for index in p.master_ndindex():
        ...     print(index)
        ...
        (3, 0, 0)
        (3, 0, 1)
        (3, 0, 2)
        (4, 0, 0)
        (4, 0, 1)
        (4, 0, 2)

        """
        return itertools_product(
            *[range(*r) for r in self.location]
        )  # TODO check

    def new_part(self, indices, master_axis_to_position, master_flip):
        """Update the `!part` attribute in-place for new indices of the
        master array.

        :Parameters:

            indices: `list`

            master_axis_to_position: `dict`

            master_flip: `list`

        :Returns:

            None

        **Examples:**

        >>> p.new_part(indices, dim2position, master_flip)

        """
        shape = self.shape

        if indices == [slice(0, stop, 1) for stop in shape]:
            return

        # ------------------------------------------------------------
        # If a dimension runs in the wrong direction then change its
        # index to account for this.
        #
        # For example, if a dimension with the wrong direction has
        # size 10 and its index is slice(3,8,2) then after the
        # direction is set correctly, the index needs to changed to
        # slice(6,0,-2):
        #
        # >>> a = [9, 8, 7, 6, 5, 4, 3, 2, 1, 0]
        # >>> a[slice(3, 8, 2)]
        # [6, 4, 2]
        # >>> a.reverse()
        # >>> print(a)
        # >>> a = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
        # >>> a[slice(6, 0, -2)]
        # [6, 4, 2]
        # ------------------------------------------------------------

        if self._subarray.size > 1:
            indices = indices[:]

            p_flip = self.flip

            for axis, i in master_axis_to_position.items():

                if (axis not in p_flip and axis not in master_flip) or (
                    axis in p_flip and axis in master_flip
                ):
                    # This axis runs in the correct direction
                    continue

                # Still here? Then this axis runs in the wrong
                # direction.

                # Reset the direction
                p_flip = p_flip[:]
                if axis in self.flip:
                    p_flip.remove(axis)
                else:
                    p_flip.append(axis)

                # Modify the index to account for the changed
                # direction
                size = shape[i]

                if isinstance(indices[i], slice):
                    start, stop, step = indices[i].indices(size)
                    # Note that step is assumed to be always +ve here
                    div, mod = divmod(stop - start - 1, step)
                    start = size - 1 - start
                    stop = start - div * step - 1
                    if stop < 0:
                        stop = None
                    indices[i] = slice(start, stop, -step)
                else:
                    size -= 1
                    indices[i] = [size - j for j in indices[i]]
            # --- End: for

            self.flip = p_flip
        # --- End: if

        slice_None = slice(None)

        # Reorder the new indices
        indices = [
            (
                indices[master_axis_to_position[axis]]
                if axis in master_axis_to_position
                else slice_None
            )
            for axis in self.axes
        ]

        part = self.part

        if not part:
            self.part = indices
            return

        # Still here? update an existing part
        p_part = []
        for part_index, index, size in zip(
            part, indices, self._subarray.shape
        ):

            if index == slice_None:
                p_part.append(part_index)
                continue

            if isinstance(part_index, slice):
                if isinstance(index, slice):

                    start, stop, step = part_index.indices(size)

                    size1, mod = divmod(stop - start - 1, step)

                    start1, stop1, step1 = index.indices(size1 + 1)

                    size2, mod = divmod(stop1 - start1, step1)

                    if mod != 0:
                        size2 += 1

                    start += start1 * step
                    step *= step1
                    stop = start + (size2 - 1) * step

                    if step > 0:
                        stop += 1
                    else:
                        stop -= 1
                    if stop < 0:
                        stop = None
                    p_part.append(slice(start, stop, step))

                    continue
                else:
                    new_part = list(range(*part_index.indices(size)))
                    new_part = [new_part[i] for i in index]
            else:
                if isinstance(index, slice):
                    new_part = part_index[index]
                else:
                    new_part = [part_index[i] for i in index]
            # --- End: if

            # Still here? Then the new element of p_part is a list of
            # integers, so let's see if we can convert it to a slice
            # before appending it.
            new_part0 = new_part[0]
            if len(new_part) == 1:
                # Convert a single element list to a slice object
                new_part = slice(new_part0, new_part0 + 1, 1)
            else:
                step = new_part[1] - new_part0
                if step:
                    if step > 0:
                        start, stop = new_part0, new_part[-1] + 1
                    else:
                        start, stop = new_part0, new_part[-1] - 1
                        if new_part == list(range(start, stop, step)):
                            if stop < 0:
                                stop = None
                            new_part = slice(start, stop, step)
            # --- End: if

            p_part.append(new_part)
        # --- End: for

        self.part = p_part

    def extra_memory(self):
        """The extra memory required to access the array."""
        if not self.in_memory:
            # --------------------------------------------------------
            # The subarray is on disk so getting the partition's data
            # array will require extra memory
            # --------------------------------------------------------
            extra_memory = True
        else:
            # --------------------------------------------------------
            # The subarray is already in memory
            # --------------------------------------------------------
            config = self.config

            p_part = self.part
            if p_part:
                extra_memory = True
            elif not config["unique_subarray"]:
                extra_memory = True
            else:
                p_data = self._subarray

                if not numpy_ma_isMA(p_data):
                    # The p_data is not a masked array
                    extra_memory = isinstance(p_data.base, numpy_ndarray)
                else:
                    # The p_data is a masked array
                    memory_overlap = isinstance(
                        p_data.data.base, numpy_ndarray
                    )
                    if not (
                        p_data.mask is numpy_ma_nomask
                        or not numpy_ma_is_masked(p_data)
                    ):
                        # There is at least one missing data point
                        memory_overlap |= isinstance(
                            p_data.mask.base, numpy_ndarray
                        )

                    extra_memory = memory_overlap
                # --- End: if

                p_dtype = p_data.dtype

                if not extra_memory:
                    if config["func"] is not None:
                        extra_memory = True
                    else:
                        p_units = self.Units
                        units = config["units"]
                        if (
                            not p_units.equals(units)
                            and bool(p_units) is bool(units)
                            and not (
                                p_data.flags["C_CONTIGUOUS"]
                                and p_dtype.kind == "f"
                            )
                        ):
                            extra_memory = True

                # ------------------------------------------------------------
                # Extra memory is required if the dtype needs changing
                # ------------------------------------------------------------
                if not extra_memory:
                    dtype = config["dtype"]
                    if dtype is not None and dtype != p_data.dtype:
                        extra_memory = True
        # --- End: if

        # ------------------------------------------------------------
        # Amount of extra memory (in bytes) required to access the
        # array
        # ------------------------------------------------------------
        return self.nbytes if extra_memory else 0

    def open(self, config):
        """Open the partition prior to getting its array.

        :Parameters:

            config: `dict`

        :Returns:

            `None`

        """
        unique_subarray = getrefcount(self._subarray) <= 2

        config = config.copy()
        config["unique_subarray"] = unique_subarray

        self.config = config

        if config.get("auxiliary_mask"):
            self._configure_auxiliary_mask(config["auxiliary_mask"])

        self.config["extra_memory"] = self.extra_memory()

        self._in_place_changes = True
        self.masked = True

        if hasattr(self, "output"):
            del self.output

        return config

    def overlaps(self, indices):
        """Return True if the subarray overlaps a subspace of the master
        array.

        :Parameters:

           indices: sequence
               Indices describing a subset of the master array. Each index
               is either a `slice` object or a `list`. If the sequence is
               empty then it is assumed that the master array is a scalar
               array.

        :Returns:

            p_indices, shape : `list`, `list` or `None`, `None`
                If the partition overlaps the *indices* then return a list
                of indices which will subset the partition's data to where
                it overlaps the master indices and the subsetted
                partition's shape as a list. Otherwise return `None`,
                `None`.

        **Examples:**

        >>> indices = (slice(None), slice(5, 1, -2), [1, 3, 4, 8])
        >>> p.overlaps(indices)
        (slice(), ddfsfsd), [3, 5, 4]

        """
        p_indices = []
        shape = []

        if not indices:
            return p_indices, shape

        for index, (r0, r1), size in zip(indices, self.location, self.shape):
            if isinstance(index, slice):
                stop = size
                if index.stop < r1:
                    stop -= r1 - index.stop

                start = index.start - r0
                if start < 0:
                    start %= index.step  # start is now +ve

                if start >= stop:
                    # This partition does not span the slice
                    return None, None

                # Still here?
                step = index.step
                index = slice(start, stop, step)
                index_size, rem = divmod(stop - start, step)
                if rem:
                    index_size += 1

            else:

                # Still here?
                index = [i - r0 for i in index if r0 <= i < r1]
                index_size = len(index)
                if index_size == 0:
                    return None, None
                elif index_size == 1:
                    index = slice(index[0], index[0] + 1)
                else:
                    index0 = index[0]
                    step = index[1] - index0
                    if step > 0:
                        start, stop = index0, index[-1] + 1
                    elif step < 0:
                        start, stop = index0, index[-1] - 1
                    if index == list(range(start, stop, step)):
                        # Replace the list with a slice object
                        if stop < 0:
                            stop = None
                        index = slice(start, stop, step)
            # --- End: if

            p_indices.append(index)
            shape.append(index_size)
        # --- End: for

        # Still here? Then this partition does span the slice and the
        # elements of this partition specified by p_indices are in the
        # slice.
        return p_indices, shape

    #    def parallelise(self, use_shared_memory=1, from_disk=True):
    #        '''
    #        '''
    #        config = self.config
    #
    #        if 1 <= use_shared_memory <= 2:
    #            self.to_shared_memory(from_disk=from_disk)
    #        elif use_shared_memory >= 3:
    #            if config['unique_subarray']:
    #                self.to_shared_memory(from_disk=from_disk)
    #
    #        if not self.in_memory or self.in_shared_memory:
    #            # Work on this partition in parallel
    #            config['serial'] = False
    #            return True
    #
    #        # Still here? Then work on this partition in serial.
    #        return False

    def to_disk(self, reopen=True):
        """Move the partition's subarray to a temporary file on disk.

        .. note:: It is assumed that the partition's subarray is currently
                  in memory, but this is not checked.

        :Parameters:

            reopen: `bool`, optional

        :Returns:

            `bool`
                True if the subarray was moved to disk, False otherwise.

        **Examples:**

        >>> p.to_disk()
        >>> p.to_disk(reopen=False)

        """
        #        try:
        tfa = CachedArray(self.array)
        #        except Exception:
        #            return False

        fd, _lock_file = mkstemp(
            prefix=tfa._partition_file + "_", dir=tfa._partition_dir
        )
        close(fd)

        self.subarray = tfa
        _temporary_files[tfa._partition_file] = (
            tfa._partition_dir,
            _lock_file,
            set(),
        )

        if reopen:
            # Re-open the partition
            self.open(self.config)

        return True

    #    def to_shared_memory(self, from_disk=True):
    #        '''
    #
    #    :Parameters:
    #
    #        from_disk : bool, optional
    #
    #    :Returns:
    #
    #        `bool`
    #            Whether or not the subarray is in shared memory.
    #
    #        '''
    #        if self.in_shared_memory:
    #            return True
    #
    #        if not from_disk and not self.on_disk:
    #            return False
    #
    #        if not self.fits_in_memory:
    #            # There is not enough space for the new shared memory
    #            # array
    #            return False
    #
    #        try:
    #            self.subarray = SharedMemoryArray(self.array)
    #        except Exception:
    #            return False
    #
    #        # Re-open the partition
    #        self.open(self.config)
    #
    #        return True

    def revert(self):
        """Completely update the partition with another partition's
        attributes in place.

        The updated partition is always dependent of the other partition.

        :Parameters:

            other: `Partition`

        :Returns:

            `None`

        **Examples:**

        >>> p.revert()

        """
        original = getattr(self, "_original", None)
        if not original:
            return

        if hasattr(self, "output"):
            output = self.output
            keep_output = True
        else:
            keep_output = False

        del self._original

        self.__dict__ = original.__dict__

        if keep_output:
            self.output = output

    def update_inplace_from(self, other):
        """Completely update the partition with another partition's
        attributes in place.

        :Parameters:

            other: `Partition`

        :Returns:

            `None`

        **Examples:**

        >>> p.update_inplace_from(q)

        """
        self.__dict__ = other.__dict__.copy()

    # def set_from_child_process(self, other):
    #     '''
    #     '''
    #     if other.in_shared_memory and self.in_shared_memory:
    #         # If BOTH subarrays are in shared memory
    #         _shared_memory_array[id(other.subarray)] = (
    #              _shared_memory_array[id(self.subarray)])
    #     if other.in_cached_file:
    #         # The other subarray is in a temporary file
    #         _temporary_files.add(other._subarray._partition_file)

    #     self.__dict__ = other.__dict__

    def _register_temporary_file(self):
        """Register a temporary file on this rank that has been created
        on another rank."""
        _partition_file = self._subarray._partition_file
        _partition_dir = self._subarray._partition_dir
        if _partition_file not in _temporary_files:
            fd, _lock_file = mkstemp(
                prefix=_partition_file + "_", dir=_partition_dir
            )
            close(fd)
            _temporary_files[_partition_file] = (
                _partition_dir,
                _lock_file,
                set(),
            )
        else:
            _, _lock_file, _ = _temporary_files[_partition_file]

        return _lock_file

    def _update_lock_files(self, lock_files):
        """Add the lock files listed in lock_files to the list of lock
        files managed by other ranks."""
        _, _lock_file, _other_lock_files = _temporary_files[
            self._subarray._partition_file
        ]
        _other_lock_files.update(set(lock_files))
        if _lock_file in _other_lock_files:
            # If the lock file managed by this rank is in the list of
            # lock files managed by other ranks, remove it from there
            _other_lock_files.remove(_lock_file)


# --- End: class



import datetime
import faulthandler
import inspect
import os
import unittest

import numpy

faulthandler.enable()  # to debug seg faults and timeouts

import cf


class Field_collapseTest(unittest.TestCase):
    def setUp(self):
        self.filename2 = os.path.join(
            os.path.dirname(os.path.abspath(__file__)), "test_file2.nc"
        )

        self.test_only = []

    #        self.test_only = ['nought']
    #        self.test_only = ['test_Field_collapse']
    #        self.test_only = ['test_Field_collapse_CLIMATOLOGICAL_TIME']
    #        self.test_only = ['test_Field_collapse_GROUPS']

    def test_Field_collapse_CLIMATOLOGICAL_TIME(self):
        if self.test_only and inspect.stack()[0][3] not in self.test_only:
            return

        verbose = False

        f = cf.example_field(2)

        g = f.collapse(
            "T: mean within years time: minimum over years",
            within_years=cf.seasons(),
        )
        expected_shape = list(f.shape)
        expected_shape[0] = 4

        if verbose:
            print("\n", f)
            print(g)
            print(g.constructs)
        self.assertEqual(list(g.shape), expected_shape, g.shape)

        g = f.collapse(
            "T: max within years time: minimum over years",
            within_years=cf.seasons(),
        )
        expected_shape = list(f.shape)
        expected_shape[0] = 4

        if verbose:
            print("\n", f)
            print(g)
            print(g.constructs)
        self.assertEqual(list(g.shape), expected_shape)

        g = f.collapse(
            "T: mean within years time: minimum over years",
            within_years=cf.M(),
        )
        expected_shape = list(f.shape)
        expected_shape[0] = 12

        if verbose:
            print("\n", f)
            print(g)
            print(g.constructs)
        self.assertEqual(list(g.shape), expected_shape)

        g = f.collapse(
            "T: max within years time: minimum over years", within_years=cf.M()
        )
        expected_shape = list(f.shape)
        expected_shape[0] = 12

        if verbose:
            print("\n", f)
            print(g)
            print(g.constructs)
        self.assertEqual(list(g.shape), expected_shape)

        g = f[:12].collapse(
            "T: mean within years time: minimum over years",
            within_years=cf.seasons(),
        )
        expected_shape = list(f.shape)
        expected_shape[0] = 4

        if verbose:
            print("\n", f[:12])
            print(g)
            print(g.constructs)
        self.assertEqual(list(g.shape), expected_shape)

        g = f[:12].collapse(
            "T: max within years time: minimum over years",
            within_years=cf.seasons(),
        )
        expected_shape = list(f.shape)
        expected_shape[0] = 4

        if verbose:
            print("\n", f[:12])
            print(g)
            print(g.constructs)
        self.assertEqual(list(g.shape), expected_shape)

        g = f[:12].collapse(
            "T: mean within years time: minimum over years",
            within_years=cf.M(),
        )
        expected_shape = list(f.shape)
        expected_shape[0] = 12

        if verbose:
            print("\n", f[:12])
            print(g)
            print(g.constructs)
        self.assertEqual(list(g.shape), expected_shape)

        g = f[:12].collapse(
            "T: max within years time: minimum over years", within_years=cf.M()
        )
        expected_shape = list(f.shape)
        expected_shape[0] = 12

        if verbose:
            print("\n", f[:12])
            print(g)
            print(g.constructs)
        self.assertEqual(list(g.shape), expected_shape)

        for key in f.cell_methods(todict=True):
            f.del_construct(key)

        g = f.collapse(
            "T: max within years time: minimum over years",
            within_years=cf.seasons(),
        )
        expected_shape = list(f.shape)
        expected_shape[0] = 4

        if verbose:
            print("\n", f)
            print(g)
            print(g.constructs)
        self.assertEqual(list(g.shape), expected_shape)

        g = f.collapse(
            "T: max within years time: min over years", within_years=cf.M()
        )
        expected_shape = list(f.shape)
        expected_shape[0] = 12

        if verbose:
            print("\n", f)
            print(g)
            print(g.constructs)
        self.assertEqual(list(g.shape), expected_shape)

        g = f[:12].collapse(
            "T: max within years time: minimum over years",
            within_years=cf.seasons(),
        )
        expected_shape = list(f.shape)
        expected_shape[0] = 4

        if verbose:
            print("\n", f[:12])
            print(g)
            print(g.constructs)
        self.assertEqual(list(g.shape), expected_shape)

        g = f[:12].collapse(
            "T: max within years time: minimum over years", within_years=cf.M()
        )
        expected_shape = list(f.shape)
        expected_shape[0] = 12

        if verbose:
            print("\n", f[:12])
            print(g)
            print(g.constructs)
        self.assertEqual(list(g.shape), expected_shape)

        g = f.collapse(
            "T: max within years time: minimum over years",
            within_years=cf.seasons(),
            over_years=cf.Y(2),
        )
        expected_shape = list(f.shape)
        expected_shape[0] = 8

        if verbose:
            print("\n", f)
            print(g)
            print(g.constructs)
        self.assertEqual(list(g.shape), expected_shape)

        g = f[::-1, ...].collapse(
            "T: max within years time: minimum over years",
            within_years=cf.seasons(),
            over_years=cf.Y(2),
        )
        expected_shape = list(f.shape)
        expected_shape[0] = 8

        if verbose:
            print("\n", f[::-1, ...])
            print(g)
            print(g.constructs)
        self.assertEqual(list(g.shape), expected_shape)

    def test_Field_collapse(self):
        if self.test_only and inspect.stack()[0][3] not in self.test_only:
            return

        verbose = False

        f = cf.read(self.filename2)[0]

        g = f.collapse("mean")
        if verbose:
            print("\n", f)
            print(g)
            print(g.constructs)
        self.assertEqual(g.shape, (1, 1, 1), g.shape)

        g = f.collapse("mean", axes=["T", "X"])
        if verbose:
            print("\n", f)
            print(g)
            print(g.constructs)
        self.assertEqual(g.shape, (1, 4, 1))

        g = f.collapse("mean", axes=[0, 2])

        if verbose:
            print("\n", f)
            print(g)
            print(g.constructs)
        self.assertEqual(g.shape, (1, 4, 1))

        g = f.collapse("mean", axes=[0, 1])
        if verbose:
            print("\n", f)
            print(g)
            print(g.constructs)
        self.assertEqual(g.shape, (1, 1, 5))

        g = f.collapse("mean", axes="domainaxis1")
        if verbose:
            print("\n", f)
            print(g)
            print(g.constructs)
        self.assertEqual(g.shape, (1800, 1, 5))

        g = f.collapse("mean", axes=["domainaxis1"])
        if verbose:
            print("\n", f)
            print(g)
            print(g.constructs)
        self.assertEqual(g.shape, (1800, 1, 5))

        g = f.collapse("mean", axes=[1])
        if verbose:
            print("\n", f)
            print(g)
            print(g.constructs)
        self.assertEqual(g.shape, (1800, 1, 5))

        g = f.collapse("mean", axes=1)
        if verbose:
            print("\n", f)
            print(g)
            print(g.constructs)
        self.assertEqual(g.shape, (1800, 1, 5))

        g = f.collapse("T: mean")
        if verbose:
            print("\n", f)
            print(g)
            print(g.constructs)
        self.assertEqual(g.shape, (1, 4, 5))

        g = f.collapse("T: mean X: maximum")
        if verbose:
            print("\n", f)
            print(g)
            print(g.constructs)
        self.assertEqual(g.shape, (1, 4, 1))

        g = f.collapse(
            "T: mean within years time: minimum over years",
            within_years=cf.M(),
        )
        if verbose:
            print("\n", f)
            print(g)
            print(g.constructs)
        self.assertEqual(g.shape, (12, 4, 5))

        for m in range(1, 13):
            a = numpy.empty((5, 4, 5))
            for i, year in enumerate(
                f.subspace(T=cf.month(m)).coord("T").year.unique()
            ):
                cf.month(m) & cf.year(year)
                x = f.subspace(T=cf.month(m) & cf.year(year))
                x.data.mean(axes=0, inplace=True)
                a[i] = x.array

            a = a.min(axis=0)
            self.assertTrue(numpy.allclose(a, g.array[m % 12]))

        g = f.collapse("T: mean", group=360)

        for group in (
            cf.M(12),
            cf.M(12, month=12),
            cf.M(12, day=16),
            cf.M(12, month=11, day=27),
        ):
            g = f.collapse("T: mean", group=group)
            bound = g.coord("T").bounds.datetime_array[0, 1]
            self.assertEqual(
                bound.month,
                group.offset.month,
                f"{bound.month}!={group.offset.month}, group={group}",
            )
            self.assertEqual(
                bound.day,
                group.offset.day,
                f"{bound.day}!={group.offset.day}, group={group}",
            )

    #            for group in (cf.D(30),
    #                          cf.D(30, month=12),
    #                          cf.D(30, day=16),
    #                          cf.D(30, month=11, day=27)):
    #                g = f.collapse('T: mean', group=group)
    #                bound = g.coord('T').bounds.datetime_array[0, 1]
    #                self.assertEqual(
    #                    bound.day, group.offset.day,
    #                    "{}!={}, bound={}, group={}".format(
    #                        bound.day, group.offset.day, bound, group)
    #                )

    def test_Field_collapse_WEIGHTS(self):
        if self.test_only and inspect.stack()[0][3] not in self.test_only:
            return

        verbose = False

        f = cf.example_field(2)

        if verbose:
            print(f)

        g = f.collapse("area: mean")
        g = f.collapse("area: mean", weights="area")
        if verbose:
            print(g)

        # Check area/volume collapses on fields with a different setup:
        h = cf.example_field(3)
        h.collapse("volume: minimum")
        i = cf.example_field(4)
        i.collapse("area: maximum")

    def test_Field_collapse_GROUPS(self):
        if self.test_only and inspect.stack()[0][3] not in self.test_only:
            return

        verbose = False

        f = cf.example_field(2)

        g = f.collapse("T: mean", group=cf.M(12), group_span=cf.Y())
        expected_shape = list(f.shape)
        expected_shape[0] = 2

        if verbose:
            print(f)
            print(g)
            print(g.constructs)
        self.assertEqual(list(g.shape), expected_shape, g.shape)

        g = f.collapse("T: mean", group=cf.M(12, month=12), group_span=cf.Y())
        expected_shape = list(f.shape)
        expected_shape[0] = 3

        if verbose:
            print(f)
            print(g)
            print(g.constructs)
        self.assertEqual(list(g.shape), expected_shape, g.shape)

        g = f.collapse("T: mean", group=cf.M(12, day=16), group_span=cf.Y())
        expected_shape = list(f.shape)
        expected_shape[0] = 2

        if verbose:
            print(f)
            print(g)
            print(g.constructs)
        self.assertEqual(list(g.shape), expected_shape, g.shape)

        g = f.collapse(
            "T: mean", group=cf.M(12, month=11, day=27), group_span=cf.Y()
        )
        expected_shape = list(f.shape)
        expected_shape[0] = 3

        if verbose:
            print(f)
            print(g)
            print(g.constructs)
        self.assertEqual(list(g.shape), expected_shape, g.shape)

        g = f.collapse(
            "T: mean", group=cf.M(12, month=6, day=27), group_span=cf.Y()
        )
        expected_shape = list(f.shape)
        expected_shape[0] = 2

        if verbose:
            print(f)
            print(g)
            print(
                g.dimension_coordinates("T").value().bounds.data.datetime_array
            )
            print(g.constructs)
        self.assertEqual(list(g.shape), expected_shape, g.shape)

        g = f.collapse(
            "T: mean",
            group=cf.M(5, month=12),
            group_span=cf.M(5),
            group_contiguous=1,
        )
        expected_shape = list(f.shape)
        expected_shape[0] = 7

        if verbose:
            print(f)
            print(g)
            print(
                g.dimension_coordinates("T").value().bounds.data.datetime_array
            )
            print(g.constructs)
        self.assertEqual(list(g.shape), expected_shape, g.shape)

        g = f.collapse(
            "T: mean",
            group=cf.M(5, month=12),
            group_span=cf.M(5),
            group_contiguous=1,
        )
        expected_shape = list(f.shape)
        expected_shape[0] = 7

        if verbose:
            print(f)
            print(g)
            print(
                g.dimension_coordinates("T").value().bounds.data.datetime_array
            )
            print(g.constructs)
        self.assertEqual(list(g.shape), expected_shape, g.shape)

        g = f.collapse(
            "T: mean",
            group=cf.M(5, month=3),
            group_span=cf.M(5),
            group_contiguous=1,
        )
        expected_shape = list(f.shape)
        expected_shape[0] = 7

        if verbose:
            print(f)
            print(g)
            print(
                g.dimension_coordinates("T").value().bounds.data.datetime_array
            )
            print(g.constructs)
        self.assertEqual(list(g.shape), expected_shape, g.shape)

        g = f.collapse(
            "T: mean",
            group=cf.M(5, month=2),
            group_span=cf.M(5),
            group_contiguous=1,
        )
        expected_shape = list(f.shape)
        expected_shape[0] = 7

        if verbose:
            print(f)
            print(g)
            print(
                g.dimension_coordinates("T").value().bounds.data.datetime_array
            )
            print(g.constructs)
        self.assertEqual(list(g.shape), expected_shape, g.shape)

        g = f.collapse(
            "T: mean",
            group=cf.M(5, month=12),
            group_span=cf.M(5),
            group_contiguous=2,
        )
        expected_shape = list(f.shape)
        expected_shape[0] = 7

        if verbose:
            print(f)
            print(g)
            print(
                g.dimension_coordinates("T").value().bounds.data.datetime_array
            )
            print(g.constructs)
        self.assertEqual(list(g.shape), expected_shape, g.shape)

        g = f.collapse("T: mean", group=cf.M(5, month=3))
        expected_shape = list(f.shape)
        expected_shape[0] = 7

        if verbose:
            print(f)
            print(g)
            print(
                g.dimension_coordinates("T").value().bounds.data.datetime_array
            )
            print(g.constructs)
        self.assertEqual(list(g.shape), expected_shape, g.shape)
        # TODO - look into month offset when M< 12

        g = f.collapse(
            "T: mean",
            group=cf.M(5, month=3),
            group_span=cf.M(5),
            group_contiguous=2,
        )
        expected_shape = list(f.shape)
        expected_shape[0] = 7

        if verbose:
            print(f)
            print(g)
            print(
                g.dimension_coordinates("T").value().bounds.data.datetime_array
            )
            print(g.constructs)
        self.assertEqual(list(g.shape), expected_shape, g.shape)

        g = f.collapse("T: mean", group=cf.M(5, month=12), group_contiguous=1)
        expected_shape = list(f.shape)
        expected_shape[0] = 7

        if verbose:
            print(f)
            print(g)
            print(
                g.dimension_coordinates("T").value().bounds.data.datetime_array
            )
            print(g.constructs)
        self.assertEqual(list(g.shape), expected_shape, g.shape)

        g = f.collapse("T: mean", group=cf.M(5, month=3), group_contiguous=1)
        expected_shape = list(f.shape)
        expected_shape[0] = 7

        if verbose:
            print(f)
            print(g)
            print(
                g.dimension_coordinates("T").value().bounds.data.datetime_array
            )
            print(g.constructs)
        self.assertEqual(list(g.shape), expected_shape, g.shape)

        g = f.collapse("T: mean", group=cf.M(5, month=12), group_contiguous=2)
        expected_shape = list(f.shape)
        expected_shape[0] = 7

        if verbose:
            print(f)
            print(g)
            print(
                g.dimension_coordinates("T").value().bounds.data.datetime_array
            )
            print(g.constructs)
        self.assertEqual(list(g.shape), expected_shape, g.shape)

        # Test method=integral with groups
        g = f.collapse(
            "T: integral", group=cf.M(5, month=12), weights=True, measure=True
        )
        expected_shape = list(f.shape)
        expected_shape[0] = 7
        self.assertEqual(list(g.shape), expected_shape, g.shape)

        g = f.collapse("T: mean", group=cf.M(5, month=3), group_contiguous=2)
        expected_shape = list(f.shape)
        expected_shape[0] = 7

        if verbose:
            print(f)
            print(g)
            print(
                g.dimension_coordinates("T").value().bounds.data.datetime_array
            )
            print(g.constructs)
        self.assertEqual(list(g.shape), expected_shape, g.shape)

        g = f.collapse(
            "T: mean within years time: minimum over years",
            within_years=cf.M(3),
            group_span=True,
        )
        expected_shape = list(f.shape)
        expected_shape[0] = 4

        if verbose:
            print(f)
            print(g)
            print(
                g.dimension_coordinates("T").value().bounds.data.datetime_array
            )
            print(g.constructs)
        self.assertEqual(list(g.shape), expected_shape, g.shape)

        g = f.collapse(
            "T: mean within years time: minimum over years",
            within_years=cf.seasons(),
            group_span=cf.M(3),
        )
        expected_shape = list(f.shape)
        expected_shape[0] = 4

        if verbose:
            print(f)
            print(g)
            print(
                g.dimension_coordinates("T").value().bounds.data.datetime_array
            )
            print(g.constructs)
        self.assertEqual(list(g.shape), expected_shape, g.shape)


#            g = f[::2].collapse('T: mean', group=cf.M(5, month=12),
#                                group_span=cf.M(5),group_contiguous=1)
#            g = f.collapse('T: mean', group=cf.M(5, month= 3),
#                           group_contiguous=1)
#            g = f.collapse('T: mean', group=cf.M(5, month=12),
#                           group_contiguous=2)
#            g = f.collapse('T: mean', group=cf.M(5, month= 3),
#                           group_contiguous=2)


if __name__ == "__main__":
    print("Run date:", datetime.datetime.now())
    cf.environment()
    print()
    unittest.main(verbosity=2)



import numpy

from ..functions import get_subspace, parse_indices
from ..umread_lib.umfile import Rec
from . import abstract
from .functions import _close_um_file, _open_um_file


class UMArray(abstract.FileArray):
    """A sub-array stored in a PP or UM fields file."""

    def __init__(
        self,
        filename=None,
        dtype=None,
        ndim=None,
        shape=None,
        size=None,
        header_offset=None,
        data_offset=None,
        disk_length=None,
        fmt=None,
        word_size=None,
        byte_ordering=None,
    ):
        """**Initialization**

        :Parameters:

            filename: `str`
                The file name in normalized, absolute form.

            dtype: `numpy.dtype`
                The data type of the data array on disk.

            ndim: `int`
                The number of dimensions in the unpacked data array.

            shape: `tuple`
                The shape of the unpacked data array.

            size: `int`
                The number of elements in the unpacked data array.

            header_offset: `int`
                The start position in the file of the header.

            data_offset: `int`
                The start position in the file of the data array.

            disk_length: `int`
                The number of words on disk for the data array, usually
                LBLREC-LBEXT. If set to 0 then `!size` is used.

            fmt: `str`, optional
                The file format of the UM file containing the array
                ('FF' or 'PP').

            word_size: `int`, optional
                Word size in bytes (4 or 8).

            byte_ordering: `str`, optional
                The endianness of the data ('little_endian' or 'big_endian').

        **Examples:**

        >>> a = UMFileArray(file='file.pp', header_offset=3156, data_offset=3420,
        ...                 dtype=numpy.dtype('float32'), shape=(30, 24),
        ...                 size=720, ndim=2, disk_length=0)

        >>> a = UMFileArray(
        ...         file='packed_file.pp', header_offset=3156, data_offset=3420,
        ...         dtype=numpy.dtype('float32'), shape=(30, 24),
        ...         size=720, ndim=2, disk_length=423
        ...     )

        """
        super().__init__(
            filename=filename,
            dtype=dtype,
            ndim=ndim,
            shape=shape,
            size=size,
            header_offset=header_offset,
            data_offset=data_offset,
            disk_length=disk_length,
            fmt=fmt,
            word_size=word_size,
            byte_ordering=byte_ordering,
        )

        # By default, do not close the UM file after data array access
        self._close = False

    def __getitem__(self, indices):
        """Implement indexing.

        x.__getitem__(indices) <==> x[indices]

        :Returns:

            `numpy.ndarray`

        """
        f = self.open()

        rec = Rec.from_file_and_offsets(
            f, self.header_offset, self.data_offset, self.disk_length
        )

        int_hdr = rec.int_hdr
        real_hdr = rec.real_hdr

        array = rec.get_data().reshape(
            int_hdr.item(
                17,
            ),
            int_hdr.item(
                18,
            ),
        )

        if indices is not Ellipsis:
            indices = parse_indices(array.shape, indices)
            array = get_subspace(array, indices)

        LBUSER2 = int_hdr.item(
            38,
        )

        if LBUSER2 == 3:
            # Return the numpy array now if it is a boolean array
            return array.astype(bool)

        integer_array = LBUSER2 == 2

        # ------------------------------------------------------------
        # Convert to a masked array
        # ------------------------------------------------------------
        # Set the fill_value from BMDI
        fill_value = real_hdr.item(
            17,
        )
        if fill_value != -1.0e30:
            # -1.0e30 is the flag for no missing data
            if integer_array:
                # The fill_value must be of the same type as the data
                # values
                fill_value = int(fill_value)

            # Mask any missing values
            mask = array == fill_value
            if mask.any():
                array = numpy.ma.masked_where(mask, array, copy=False)
        # --- End: if

        # ------------------------------------------------------------
        # Unpack the array using the scale_factor and add_offset, if
        # either is available
        # ------------------------------------------------------------
        # Treat BMKS as a scale_factor if it is neither 0 nor 1
        scale_factor = real_hdr.item(
            18,
        )
        if scale_factor != 1.0 and scale_factor != 0.0:
            if integer_array:
                scale_factor = int(scale_factor)
            array *= scale_factor

        # Treat BDATUM as an add_offset if it is not 0
        add_offset = real_hdr.item(
            4,
        )
        if add_offset != 0.0:
            if integer_array:
                add_offset = int(add_offset)
            array += add_offset

        # Return the numpy array
        return array

    def __str__(self):
        """x.__str__() <==> str(x)

        :Returns:

            `str`

        """
        return "%s%s in %s" % (self.header_offset, self.shape, self.filename)

    @property
    def file_pointer(self):
        """The file pointer starting at the position of the header.

        :Returns:

            2-`tuple`

        """
        return (self.filename, self.header_offset)

    @property
    def header_offset(self):
        """The start position in the file of the header.

        :Returns:

            `int`

        """
        return self._get_component("header_offset")

    @property
    def data_offset(self):
        """The start position in the file of the data array.

        :Returns:

            `int`

        """
        return self._get_component("data_offset")

    @property
    def disk_length(self):
        """The number of words on disk for the data array.

        :Returns:

            `int`

        """
        return self._get_component("disk_length")

    @property
    def fmt(self):
        """The file format of the UM file containing the array.

        :Returns:

            `str`
                'FF' or 'PP'

        """
        return self._get_component("fmt")

    @property
    def byte_ordering(self):
        """The endianness of the data.

        :Returns:

            `str`
                'little_endian' or 'big_endian'

        """
        return self._get_component("byte_ordering")

    @property
    def word_size(self):
        """Word size in bytes.

        :Returns:

            `int`
                4 or 8

        """
        return self._get_component("word_size")

    def close(self):
        """Close the file containing the data array.

        If the file is not open then no action is taken.

        :Returns:

            `None`

        **Examples:**

        >>> f.close()

        """
        _close_um_file(self.filename)

    def open(self):
        """Open the file containing the data array.

        :Returns:

            `umfile_lib.File`

        **Examples:**

        >>> f.open()

        """
        return _open_um_file(
            self.filename,
            fmt=self.fmt,
            word_size=self.word_size,
            byte_ordering=self.byte_ordering,
        )



import datetime
import faulthandler
import unittest

import numpy as np

faulthandler.enable()  # to debug seg faults and timeouts

import cf
from cf import Units


class DatetimeTest(unittest.TestCase):
    def test_Datetime(self):
        cf.dt(2003)
        cf.dt(2003, 2)
        cf.dt(2003, 2, 30, calendar="360_day")
        cf.dt(2003, 2, 30, 0, 0, calendar="360_day")
        cf.dt(2003, 2, 30, 0, 0, 0, calendar="360_day")
        cf.dt(2003, 4, 5, 12, 30, 15)
        cf.dt(2003, month=4, day=5, hour=12, minute=30, second=15)

    def test_Datetime_rt2dt(self):
        for a in (1, np.array(1), np.ma.array(1)):
            self.assertEqual(
                cf.cfdatetime.rt2dt(a, Units("days since 2004-2-28")),
                np.array(cf.dt(2004, 2, 29, calendar="standard"), dtype="O"),
            )

        for a in (np.ma.array(1, mask=True), np.ma.array([1], mask=True)):
            b = cf.cfdatetime.rt2dt(a, Units("days since 2004-2-28"))
            self.assertIsInstance(b, np.ndarray)
            self.assertEqual(b.mask, True)

        self.assertTrue(
            (
                cf.cfdatetime.rt2dt([1, 3], Units("days since 2004-2-28"))
                == np.array(
                    [
                        datetime.datetime(2004, 2, 29),
                        datetime.datetime(2004, 3, 2),
                    ]
                )
            ).all()
        )

        a = np.array(
            [
                cf.dt(2004, 2, 29, calendar=None),
                cf.dt(2004, 3, 2, calendar="gregorian"),
            ],
            dtype="O",
        )
        b = cf.cfdatetime.rt2dt([1, 3], Units("days since 2004-2-28"))
        self.assertTrue((a == b).all())

        for a in (
            np.ma.array(3),
            np.ma.array([3]),
        ):
            b = cf.cfdatetime.rt2dt(a, Units("days since 1970-01-01"))
            self.assertEqual(b, cf.dt(1970, 1, 4, calendar="gregorian"))

        for a in (
            np.ma.array(3, mask=True),
            np.ma.array([3], mask=True),
        ):
            b = cf.cfdatetime.rt2dt(a, Units("days since 1970-01-01"))
            self.assertEqual(b.mask, True)

    def test_Datetime_dt2rt(self):
        units = Units("days since 2004-2-28")
        self.assertEqual(
            cf.cfdatetime.dt2rt(datetime.datetime(2004, 2, 29), None, units),
            np.array(1.0),
        )
        self.assertTrue(
            (
                cf.cfdatetime.dt2rt(
                    [
                        datetime.datetime(2004, 2, 29),
                        datetime.datetime(2004, 3, 2),
                    ],
                    None,
                    units,
                )
                == np.array([1.0, 3.0])
            ).all()
        )
        units = Units("days since 2004-2-28", "360_day")
        self.assertTrue(
            (
                cf.cfdatetime.dt2rt(
                    [cf.dt(2004, 2, 29), cf.dt(2004, 3, 1)], None, units
                )
                == np.array([1.0, 3.0])
            ).all()
        )
        units = Units("seconds since 2004-2-28")
        self.assertEqual(
            cf.cfdatetime.dt2rt(datetime.datetime(2004, 2, 29), None, units),
            np.array(86400.0),
        )

    def test_Datetime_Data(self):
        d = cf.Data([1, 2, 3], "days since 2004-02-28")
        self.assertTrue((d < cf.dt(2005, 2, 28)).all())
        with self.assertRaises(Exception):
            d < cf.dt(2005, 2, 29)

        with self.assertRaises(Exception):
            d < cf.dt(2005, 2, 29, calendar="360_day")

        d = cf.Data([1, 2, 3], "days since 2004-02-28", calendar="360_day")
        self.assertTrue((d < cf.dt(2005, 2, 28)).all())
        self.assertTrue((d < cf.dt(2005, 2, 29)).all())
        self.assertTrue((d < cf.dt(2005, 2, 30)).all())

        with self.assertRaises(Exception):
            d < cf.dt(2005, 2, 31)

        with self.assertRaises(Exception):
            d < cf.dt(2005, 2, 29, calendar="noleap")

    def test_Datetime_dt_vector(self):
        for v in (2000, [2000], [[2000]], "2000-01-1", ["2000-01-1"]):
            x = cf.dt_vector(v)
            self.assertIsInstance(x, np.ndarray)
            self.assertEqual(x[0], cf.dt(2000, 1, 1))

        for v in ([2000, 2001], [[2000], [2001]]):
            x = cf.dt_vector(v)
            self.assertIsInstance(x, np.ndarray)
            self.assertEqual(
                x.tolist(), [cf.dt(2000, 1, 1), cf.dt(2001, 1, 1)]
            )

        for v in ([[2000, 1], [2001, 2]], ["2000-01-1", "2001-02-1"]):
            x = cf.dt_vector(v)
            self.assertIsInstance(x, np.ndarray)
            self.assertEqual(
                x.tolist(), [cf.dt(2000, 1, 1), cf.dt(2001, 2, 1)]
            )

    def test_Datetime_st2dt(self):
        for a in (
            "1970-01-04",
            np.array("1970-01-04"),
            np.ma.array(["1970-01-04"]),
        ):
            b = cf.cfdatetime.st2rt(
                a,
                Units("days since 1970-01-01"),
                Units("days since 1970-01-01"),
            )
            self.assertIsInstance(b, np.ndarray)
            self.assertEqual(b, 3)


if __name__ == "__main__":
    print("Run date:", datetime.datetime.now())
    cf.environment()
    print()
    unittest.main(verbosity=2)



from ...functions import inspect as cf_inspect
from .array import Array


class FileArray(Array):
    """A sub-array stored in a file.

    .. note:: Subclasses must define the following methods:
    `!__getitem__`, `!__str__`, `!close` and `!open`.

    """

    def __getitem__(self, indices):
        """TODO."""
        pass

    def __str__(self):
        """x.__str__() <==> str(x)"""
        return "%s in %s" % (self.shape, self.file)

    # ----------------------------------------------------------------
    # Attributes
    # ----------------------------------------------------------------
    @property
    def dtype(self):
        """Data-type of the data elements.

        **Examples:**

        >>> a.dtype
        dtype('float64')
        >>> print(type(a.dtype))
        <type 'numpy.dtype'>

        """
        return self._get_component("dtype")

    @property
    def ndim(self):
        """Number of array dimensions.

        **Examples:**

        >>> a.shape
        (73, 96)
        >>> a.ndim
        2
        >>> a.size
        7008

        >>> a.shape
        (1, 1, 1)
        >>> a.ndim
        3
        >>> a.size
        1

        >>> a.shape
        ()
        >>> a.ndim
        0
        >>> a.size
        1

        """
        return self._get_component("ndim")

    @property
    def shape(self):
        """Tuple of array dimension sizes.

        **Examples:**

        >>> a.shape
        (73, 96)
        >>> a.ndim
        2
        >>> a.size
        7008

        >>> a.shape
        (1, 1, 1)
        >>> a.ndim
        3
        >>> a.size
        1

        >>> a.shape
        ()
        >>> a.ndim
        0
        >>> a.size
        1

        """
        return self._get_component("shape")

    @property
    def size(self):
        """Number of elements in the array.

        **Examples:**

        >>> a.shape
        (73, 96)
        >>> a.size
        7008
        >>> a.ndim
        2

        >>> a.shape
        (1, 1, 1)
        >>> a.ndim
        3
        >>> a.size
        1

        >>> a.shape
        ()
        >>> a.ndim
        0
        >>> a.size
        1

        """
        return self._get_component("size")

    @property
    def filename(self):
        """The name of the file containing the array.

        **Examples:**

        >>> a.filename()
        'file.nc'

        """
        return self._get_component("filename")

    @property
    def array(self):
        """Return an independent numpy array containing the data.

        :Returns:

            `numpy.ndarray`
                An independent numpy array of the data.

        **Examples:**

        >>> n = numpy.asanyarray(a)
        >>> isinstance(n, numpy.ndarray)
        True

        """
        return self[...]

    def inspect(self):
        """Inspect the object for debugging.

        .. seealso:: `cf.inspect`

        :Returns:

            `None`

        """
        print(cf_inspect(self))  # pragma: no cover

    def get_filename(self):
        """Return the name of the file containing the array.

        **Examples:**

        >>> a.get_filename()
        'file.nc'

        """
        return self._get_component("filename")

    def close(self):
        pass

    def open(self):
        pass
